from datetime import datetime, timedelta
import pytz
from typing import List, Union, Dict
def convert_to_date(date_str):
    try:
        return datetime.strptime(date_str, '%Y-%m-%d').date()
    except ValueError:
        # Handle the error or return None if the string cannot be converted
        return 
def flatten(item, parent_key='', separator='_'):
    if item is not None:
        items = {}
        if isinstance(item, dict):
            for k, v in item.items():
                new_key = f"{parent_key}{separator}{k}" if parent_key else k
                if isinstance(v, dict):
                    items.update(flatten(v, new_key, separator=separator))
                elif isinstance(v, list):
                    for i, elem in enumerate(v):
                        items.update(flatten(elem, f"{new_key}_{i}", separator=separator))
                else:
                    items[new_key] = v
        elif isinstance(item, list):
            for i, elem in enumerate(item):
                items.update(flatten(elem, f"{parent_key}_{i}", separator=separator))
        else:
            items[parent_key] = item
        return items
def flatten_list_of_dicts(lst: List[Union[Dict, List]]) -> List[Dict]:
    return [flatten(item) for item in lst] if lst is not None else None
def flatten_dict(d, parent_key='', sep='.'):
    if d is not None:
        items = {}
        for k, v in d.items():
            new_key = f"{parent_key}{sep}{k}" if parent_key else k
            if isinstance(v, dict):
                items.update(flatten_dict(v, new_key, sep=sep))
            else:
                items[new_key] = v
        return items
async def parse_most_active(ticker_entry):

    all_parsed_data = []
    # Parsing 'ticker' attributes
    datas = ticker_entry.get('data', {})
    for data in datas:
        parsed_data = {}
        ticker_info = data.get('ticker', {})
        parsed_data['tickerId'] = ticker_info.get('tickerId')
        parsed_data['exchangeId'] = ticker_info.get('exchangeId')
        parsed_data['regionId'] = ticker_info.get('regionId')
        parsed_data['currencyId'] = ticker_info.get('currencyId')
        parsed_data['currencyCode'] = ticker_info.get('currencyCode')
        parsed_data['name'] = ticker_info.get('name')
        parsed_data['symbol'] = ticker_info.get('symbol')
        parsed_data['disSymbol'] = ticker_info.get('disSymbol')
        parsed_data['disExchangeCode'] = ticker_info.get('disExchangeCode')
        parsed_data['status'] = ticker_info.get('status')
        parsed_data['close'] = ticker_info.get('close')
        parsed_data['change'] = ticker_info.get('change')
        parsed_data['changeRatio'] = ticker_info.get('changeRatio')
        parsed_data['marketValue'] = ticker_info.get('marketValue')
        parsed_data['volume'] = ticker_info.get('volume')
        parsed_data['turnoverRate'] = ticker_info.get('turnoverRate')
        parsed_data['regionName'] = ticker_info.get('regionName')
        parsed_data['peTtm'] = ticker_info.get('peTtm')
        parsed_data['timeZone'] = ticker_info.get('timeZone')
        parsed_data['preClose'] = ticker_info.get('preClose')
        parsed_data['fiftyTwoWkHigh'] = ticker_info.get('fiftyTwoWkHigh')
        parsed_data['fiftyTwoWkLow'] = ticker_info.get('fiftyTwoWkLow')
        parsed_data['open'] = ticker_info.get('open')
        parsed_data['high'] = ticker_info.get('high')
        parsed_data['low'] = ticker_info.get('low')
        parsed_data['vibrateRatio'] = ticker_info.get('vibrateRatio')
        

        all_parsed_data.append(parsed_data)
    return all_parsed_data



# Creating a function to parse each attribute of the data_entry and return it as a dictionary
async def parse_total_top_options(data_entry):
    all_parsed_data = []
    
    for data in data_entry:
        parsed_data = {}
        ticker_info = data.get('ticker', {})
        for key, value in ticker_info.items():
            if type(key) != list and key != 'exchangeTrade' and key != 'derivativeSupport':
                parsed_data[f'{key}'] = value
    
        # Parsing 'values' attributes
        values_info = data.get('values', {})
        for key, value in values_info.items():

            if type(key) != list and key != 'exchangeTrade' and key != 'derivativeSupport':
                parsed_data[f'{key}'] = value
        
        all_parsed_data.append(parsed_data)
    if 't_sectype' in all_parsed_data:
        all_parsed_data.remove('t_sectype')

    return all_parsed_data



async def parse_contract_top_options(data_entry):
    all_parsed_data = []
    for data in data_entry:
        parsed_data = {}
        # Parsing 'belongTicker' attributes
        belong_ticker_info = data.get('belongTicker', {})
        
        for key, value in belong_ticker_info.items():
            if type(key) != list and key != 'exchangeTrade' and key != 'derivativeSupport':
                parsed_data[f'{key}'] = value

       
        
        # Parsing 'derivative' attributes
        derivative_info = data.get('derivative', {})
        for key, value in derivative_info.items():
            if type(key) != list and key != 'exchangeTrade' and key != 'derivativeSupport':
                parsed_data[f'{key}'] = value
        
        # Parsing 'values' attributes
        values_info = data.get('values', {})
        for key, value in values_info.items():
            if type(key) != list and key != 'exchangeTrade' and key != 'derivativeSupport':
                parsed_data[f'{key}'] = value 

        all_parsed_data.append(parsed_data)
    if 'bt_secType' in all_parsed_data:
        all_parsed_data = all_parsed_data.remove('bt_secType')
    return all_parsed_data



# Creating a function to parse each attribute of the data_entry and return it as a dictionary
async def parse_ticker_values(data_entry):
    all_parsed_data = []
    data_entry = data_entry.get('data', {})
    for data in data_entry:
        parsed_data = {}
        ticker_info = data.get('ticker', {})
        for key, value in ticker_info.items():
            if type(key) != list and key != 'secType' and key != 'derivativeSupport':
                parsed_data[f'{key}'] = value
    
        # Parsing 'values' attributes
        values_info = data.get('values', {})
        for key, value in values_info.items():
            if type(key) != list and key != 'secType' and key != 'derivativeSupport':
                parsed_data[f'{key}'] = value
        
        all_parsed_data.append(parsed_data)
    return all_parsed_data


def parse_forex(ticker_list):
    parsed_data_list = []
    
    for ticker_entry in ticker_list:
        parsed_data = {}
        
        parsed_data['tickerId'] = ticker_entry.get('tickerId')
        parsed_data['exchangeId'] = ticker_entry.get('exchangeId')
        
        parsed_data['name'] = ticker_entry.get('name')
        parsed_data['symbol'] = ticker_entry.get('symbol')
        parsed_data['disSymbol'] = ticker_entry.get('disSymbol')
        parsed_data['status'] = ticker_entry.get('status')
        parsed_data['close'] = ticker_entry.get('close')
        parsed_data['change'] = ticker_entry.get('change')
        parsed_data['changeRatio'] = ticker_entry.get('changeRatio')
        parsed_data['marketValue'] = ticker_entry.get('marketValue')
        
        parsed_data_list.append(parsed_data)
    
    return parsed_data_list


async def parse_etfs(response):
    flattened_data = []
    
    for tab in response.get('tabs', []):
        tab_info = {
            'id': tab.get('id'),
            'name': tab.get('name'),
            'comment': tab.get('comment'),
            'queryId': tab.get('queryId'),
            'upNum': tab.get('upNum'),
            'dowoNum': tab.get('dowoNum'),
            'flatNum': tab.get('flatNum'),
        }
        
        for ticker in tab.get('tickerTupleList', []):
            # Merge the 'tab' info and the 'ticker' info into a single dictionary
            merged_info = {**tab_info, **ticker}
            flattened_data.append(merged_info)

    return flattened_data

# Define a function to parse the given data object with specific attributes under the parent key "item"
async def parse_ipo_data(data):
    """
    Parses an IPO data object and returns a dictionary with relevant fields.

    Args:
    - item (dict): The IPO data item to parse.

    Returns:
    - dict: A dictionary containing parsed IPO data.
    """
    items = data['items']
    all_parsed_data=[]
    for item in items:
        parsed_data = {
            'ticker_id': item.get('tickerId', None),
            'list_date': item.get('listDate', None),
            'issue_up_limit': item.get('issueUpLimit', None),
            'issue_price': item.get('issuePrice', None),
            'currency_id': item.get('currencyId', None),
            'exchange_code': item.get('disExchangeCode', None),
            'symbol': item.get('disSymbol', None),
            'ipo_status': item.get('ipoStatus', None),
            'issue_currency_id': item.get('issueCurrencyId', None),
            'issue_down_limit': item.get('issueDownLimit', None),
            'issue_price_str': item.get('issuePriceStr', None),
            'name': item.get('name', None),
            'offering_type': item.get('offeringType', None),
            'prospectus': item.get('prospectus', None),
            'prospectus_publish_date': item.get('prospectusPublishDate', None),
            'purchase_end_date': item.get('purchaseEndDate', None),
            'purchase_start_date': item.get('purchaseStartDate', None),
            'close_days': item.get('closeDays', 0)  # Assuming 0 if not present
        }
        all_parsed_data.append(parsed_data)
    return all_parsed_data




# Function to convert Unix timestamps in seconds to Eastern Time in milliseconds
def convert_seconds_to_ms_eastern_time(seconds_timestamp):
    et_offset = -5 * 3600  # Eastern Standard Time (EST) offset in seconds
    utc_time = datetime.utcfromtimestamp(int(seconds_timestamp))
    eastern_time = utc_time + timedelta(seconds=et_offset)
    eastern_time_ms = int(eastern_time.timestamp() * 1000)  # Convert to milliseconds
    return eastern_time_ms


def convert_unix_to_eastern(unix_timestamp):
    eastern_time = datetime.fromtimestamp(unix_timestamp).strftime('%Y-%m-%d %H:%M:%S')
    return eastern_time


def format_date(input_str):
    # Parse the input string as a datetime object
    input_datetime = datetime.fromisoformat(input_str.replace("Z", "+00:00"))

    # Convert the datetime object to Eastern Time
    utc_timezone = pytz.timezone("UTC")
    eastern_timezone = pytz.timezone("US/Eastern")
    input_datetime = input_datetime.astimezone(utc_timezone)
    eastern_datetime = input_datetime.astimezone(eastern_timezone)

    # Format the output string
    output_str = eastern_datetime.strftime("%Y-%m-%d at %I:%M%p %Z")
    return output_str




def calculate_setup(df):
    setup_count = 0
    for i in range(3, len(df)):
        if df['Close'][i] > df['Close'][i-3]:  # Assuming 'c' is the close price column
            setup_count += 1
        else:
            setup_count = 0
        
        if setup_count >= 9:
            return True
    return False
def calculate_countdown(df):
    countdown_count = 0
    for i in range(4, len(df)):
        if df['High'][i] > df['High'][i-2]:  # Assuming 'h' is the high price column
            countdown_count += 1
        else:
            countdown_count = 0
        
        if countdown_count >= 9:
            return True


import pandas as pd

class NewsItem:
    def __init__(self, news_json):
        try:
            self.id = news_json['id'] if 'id' in news_json else None
            self.title = [i['title'] if 'title' in i else None for i in news_json]
            self.source_name = [i['sourceName'] if 'sourceName' in i else None for i in news_json]
            self.news_time = [i['newsTime']if 'newsTime' in i else None for i in news_json]
            self.news_url = [i['newsUrl'] if 'newsUrl' in i else None for i in news_json]
            self.site_type = [i['siteType'] if 'siteType' in i else None for i in news_json]
            self.collect_source = [i['collectSource']if 'collectSource' in i else None for i in news_json]



            self.data_dict = { 

                'id': self.id,
                'title': self.title,
                'source_name': self.source_name,
                'news_time': self.news_time,
                'news_url': self.news_url,
                'site_type': self.site_type,
                'collect_source': self.collect_source
            }


            self.df = pd.DataFrame(self.data_dict)
        except KeyError:
            print(f"Error getting source name for {self.news_url}")

    def __repr__(self):
        return f"NewsItem(id={self.id}, title={self.title}, source_name={self.source_name})"


import os
from dotenv import load_dotenv
load_dotenv()

from fudstop.apis.webull.webull_options import WebullOptions
from fudstop.apis.helpers import get_human_readable_string, rename_keys
import asyncpg


wb = WebullOptions(os.environ.get('WEBULL_OPTIONS'))



import asyncio
async def collect_ticker_symbol_pairs(generator):
    ticker_symbol_dict = {}
    async for ticker_id, symbol in generator:
        ticker_symbol_dict[ticker_id] = symbol
    return ticker_symbol_dict

# Assume 'self' has the 'get_ticker_symbol_pairs' method
async def main():
    await wb.connect()
    ticker_symbol_pairs_generator = wb.get_ticker_symbol_pairs()
    ticker_symbol_dict = await collect_ticker_symbol_pairs(ticker_symbol_pairs_generator)
    # Now ticker_symbol_dict contains all the ticker_id and symbol pairs

    for k,v in ticker_symbol_dict.items():
        vol_anal = await wb.fetch_volume_analysis(k)
        
        data_dict = vol_anal.data_dict
        try:
            trades_and_dates = vol_anal.flattened_trade

            print(trades_and_dates)
        except AttributeError:
            continue


        components = get_human_readable_string(f"O:{v}")
        print(components)

        data_dict.update({"option_symbol": f"{v}", "symbol": f"{components.get('underlying_symbol')}", "strike_price": f"{float(components.get('strike_price'))}", "call_put": f"{components.get('call_put')}", "expiry_date": f"{components.get('expiry_date')}"})
        trades_and_dates.update({"option_symbol": f"{v}", "symbol": f"{components.get('underlying_symbol')}", "strike_price": f"{float(components.get('strike_price'))}", "call_put": f"{components.get('call_put')}", "expiry_date": f"{components.get('expiry_date')}"})

        print(trades_and_dates)
        # Define your key mapping 
        key_mapping = {
            'tickerId': 'option_id',
            'belongTickerId': 'ticker_id',
            'totalNum': 'total_trades',
            'totalVolume': 'total_volume',
            'avgPrice': 'average_price',
            'buyVolume': 'buy_volume',
            'sellVolume': 'sell_volume',
            'neutralVolume': 'neutral_volume',
           
        }

        # Rename the keys
        renamed_data = rename_keys(data_dict, key_mapping)
        try:
            await wb.insert_trades_and_dates(trades_and_dates)
            await wb.insert_volume_analysis_data(renamed_data)
        except asyncpg.exceptions.NotNullViolationError:
            print(f'Null detected. Skipping insert.')
 
        


        


# Run the main function in the asyncio event loop
import asyncio
# assuming 'self' is an instance of the class containing the get_ticker_symbol_pairs method
asyncio.run(main())



from apis.nasdaq.nasdaq_sdk import Nasdaq


nasdaq = Nasdaq()


nasdaq.economic_indicators()


nasdaq.bigmac()


nasdaq.inflation()


nasdaq.income_expenditures()

nasdaq.unemployment()


nasdaq.sp500()


nasdaq.repo()


nasdaq.gdp()

nasdaq.interest_rates()


def kelvin_to_temp(kelvin_temp, unit="C"):
    if unit == "F":
        return (kelvin_temp - 273.15) * 9/5 + 32
    elif unit == "C":
        return kelvin_temp - 273.15
    else:
        raise ValueError("Invalid unit. Please use 'C' for Celsius or 'F' for Fahrenheit.")

# Using the function:


import os
from scripty import fetch_results,create_dataframe
import disnake
import base64
from disnake.ext import commands
from list_sets.ticker_lists import most_active_tickers
most_active_tickers = set(most_active_tickers)
from bot_menus.modals.options_modal import OptionsDataModal
import asyncio
from bot_menus.pagination import AlertMenus
import aiohttp
import matplotlib.pyplot as plt
import io
import asyncpg
import pandas as pd
from cogs.database import MyModal
from apis.polygonio.polygon_options import PolygonOptions
from apis.webull.opt_modal import OptionModal, SQLQueryModal
from apis.gexbot.gexbot import GEXBot
from dalle_modal import DalleModal
opts = PolygonOptions()
gexbot = GEXBot()
bot = commands.Bot(command_prefix="!", intents=disnake.Intents.all())

from cogs.database import QueryView
from list_sets.ticker_lists import gex_tickers
from typing import List
import disnake
from disnake.ext import commands
import openai

import os
from dotenv import load_dotenv
load_dotenv()
# Initialize the OpenAI client with your API key
openai.api_key = os.getenv("YOUR_OPENAI_KEY")

from openai import OpenAI

@bot.event
async def on_message(message: disnake.Message):
    """Use  GPT4 Vision to listen for image URLs"""

    if message.channel.id == 896207280117264434:


        print(message.content)
        if message.content.endswith('.png'):

            response = client.chat.completions.create(
                model="gpt-4-vision-preview",
                messages=[
                    {
                        "role": "user",
                        "content": [
                            {"type": "text", "text": f"Explain the options chain to the user. You're an option trader expert. Call open interest = resistance. Put open interest = support. Look for call skews - or lower iV out of the money. Take note of volume values, open interest values - reiterate them back to the user."},
                            {
                                "type": "image_url",
                                "image_url": message.content,
                            },
                        ],
                    }
                ],
                max_tokens=2000,
            )

            await message.channel.send(response.choices[0].message.content[:2000])
 
        if message.content.endswith('.jpg'):

            response = client.chat.completions.create(
                model="gpt-4-vision-preview",
                messages=[
                    {
                        "role": "user",
                        "content": [
                            {"type": "text", "text": "What’s in this image?"},
                            {
                                "type": "image_url",
                                "image_url": message.content,
                            },
                        ],
                    }
                ],
                max_tokens=300,
            )

            await message.channel.send(response.choices[0].message)

    await bot.process_commands(message)
# This dictionary will hold the conversation state for each user
conversations = {}
client = OpenAI(api_key=os.environ.get('YOUR_OPENAI_KEY'))
@bot.slash_command()
async def option_data(inter: disnake.ApplicationCommandInteraction):
    modal = OptionModal()
    await inter.response.send_modal(modal)



@bot.slash_command()
async def options_database(inter:disnake.AppCmdInter):
    """Query options data"""
    await inter.response.send_modal(OptionsDataModal())


# @bot.slash_command()
# async def options_database(inter:disnake.ApplicationCommandInteraction):
#     """Use a Modal to query the database for options."""
#     modal = SQLQueryModal()
#     await inter.response.send_modal(modal)
@bot.command()
async def turbo(ctx):
    # Start a new conversation with the user
    conversations[ctx.author.id] = []
    
    # Send an initial message to the user
    await ctx.send("> # Slave Bot\n> Online. \n\n> Your work is my...work.. Type to chat... or.. type stop to quit.")

    while True:
        # Wait for a message from the same user
        message = await bot.wait_for(
            "message",
            check=lambda m: m.author == ctx.author and m.channel == ctx.channel
        )

        # Check if the user wants to stop the conversation
        if message.content.lower() == "stop":
            await ctx.send("Goodbye! If you need help again, just call me.")
            del conversations[ctx.author.id]  # Clean up the conversation
            break

        # Append the user's message to the conversation
        conversations[ctx.author.id].append({"role": "user", "content": message.content + "YOU ARE ONLY TO REPLY IN CODE. CODE ONLY. NO MARKDOWN. ONLY CODE!"})

        # Send the conversation to OpenAI
        response = client.chat.completions.create(
            model="gpt-4-1106-preview",
            messages=conversations[ctx.author.id],
            temperature=0.33,
            max_tokens=4096        )

        # Get the content from the OpenAI response
        ai_response = response.choices[0].message.content

        # Append the AI's response to the conversation
        conversations[ctx.author.id].append({"role": "assistant", "content": ai_response})
        embeds = []
        chunks = [ai_response[i:i + 3860] for i in range(0, len(ai_response), 3860)]
        for chunk in chunks:
            embed = disnake.Embed(title=f"GPT4-Turbo", description=f"{chunk}", color=disnake.Colour.dark_orange())
            embed.add_field(name=f"Your prompt:", value=f"> **{message.content[:400]}**", inline=False)
            embeds.append(embed)

        await ctx.send(embed=embeds[0], view=AlertMenus(embeds))

@bot.slash_command()
async def dalle3(inter: disnake.AppCmdInter):
    """Generates an image using openai's Dalle3"""
    await inter.response.send_modal(modal=DalleModal(bot))


@bot.command()
async def cfr(ctx, *, query: str):
    results = fetch_results(query)
    
    async def create_table(conn):
        await conn.execute('''
            CREATE TABLE IF NOT EXISTS cfr_data (
                starts_on TEXT,
                ends_on TEXT,
                type TEXT,
                hierarchy_title TEXT,
                hierarchy_subtitle TEXT,
                hierarchy_chapter TEXT,
                hierarchy_subchapter TEXT,
                hierarchy_part TEXT,
                hierarchy_subpart TEXT,
                hierarchy_subject_group TEXT,
                hierarchy_section TEXT,
                hierarchy_appendix TEXT,
                hierarchy_headings_title TEXT,
                hierarchy_headings_subtitle TEXT,
                hierarchy_headings_chapter TEXT,
                hierarchy_headings_subchapter TEXT,
                hierarchy_headings_part TEXT,
                hierarchy_headings_subpart TEXT,
                hierarchy_headings_subject_group TEXT,
                hierarchy_headings_section TEXT,
                hierarchy_headings_appendix TEXT,
                headings_title TEXT,
                headings_subtitle TEXT,
                headings_chapter TEXT,
                headings_subchapter TEXT,
                headings_part TEXT,
                headings_subpart TEXT,
                headings_subject_group TEXT,
                headings_section TEXT,
                headings_appendix TEXT,
                full_text_excerpt TEXT,
                score REAL,
                structure_index INT,
                reserved BOOLEAN,
                removed BOOLEAN,
                change_types_effective_cross_reference TEXT,
                change_types_cross_reference TEXT,
                change_types_effective TEXT,
                change_types_initial TEXT
            );
        ''')

    async def insert_data(conn, results):
        insert_query = '''
            INSERT INTO cfr_data (
                starts_on, ends_on, type, hierarchy_title, hierarchy_subtitle,
                hierarchy_chapter, hierarchy_subchapter, hierarchy_part,
                hierarchy_subpart, hierarchy_subject_group, hierarchy_section,
                hierarchy_appendix, hierarchy_headings_title,
                hierarchy_headings_subtitle, hierarchy_headings_chapter,
                hierarchy_headings_subchapter, hierarchy_headings_part,
                hierarchy_headings_subpart, hierarchy_headings_subject_group,
                hierarchy_headings_section
            ) VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, $20, $21, $22, $23, $24, $25, $26, $27, $28, $29, $30, $31, $32, $33, $34, $35, $36, $37, $38);
        '''
        await conn.executemany(insert_query, results)
    
    if 'results' in results:

        df = pd.DataFrame(results['results'])
        csv_buffer = io.StringIO()
        df.to_csv(csv_buffer, index=False)
        csv_buffer.seek(0)
        await ctx.send(file=disnake.File(fp=csv_buffer, filename='cfr_results.csv'))
    





    


        # # Validate tickers
        # valid_tickers = [ticker for ticker in tickers if ticker in self.valid_tickers]
        # invalid_tickers = list(set(tickers) - set(valid_tickers))

        # if not valid_tickers:
        #     # No valid tickers were entered
        #     await interaction.response.send_message(
        #         "None of the entered tickers were recognized. Please try again with valid ticker symbols.",
        #         ephemeral=True
        #     )
        #     return

        # # If there are invalid tickers, inform the user but proceed with valid ones
        # if invalid_tickers:
        #     await interaction.followup.send(
        #         f"The following tickers were not recognized and will be ignored: {', '.join(invalid_tickers)}",
        #         ephemeral=True
        #     )

        # # Proceed with scanning the valid tickers
        # results = await scan_bars(valid_tickers, timeframe)
        # await interaction.followup.send(results)

from live_markets.stock_market import StockMarketLive




stock_market = StockMarketLive()
@bot.slash_command()
async def stream(inter: disnake.AppCmdInter, ticker:str):
    """Stream live trades for a ticker."""
    await inter.response.defer()
    
    await stock_market.connect()
    counter = 0
    while True:
        counter = counter + 1
        data = await stock_market.fetch_latest_trade(ticker)
        if data:
            # Format the message with the trade data
            message = f"> # Latest trade for {ticker} | Price: ${data['price']} | Size: {data['size']} | Time: {data['timestamp']}"
        else:
            # No trade data found
            message = f"> No recent trades found for {ticker}."
        await inter.edit_original_message(f"> # {message}")

        if counter == 250:
            await inter.send(f'> # Stream ended.')
            break
    





timeframe_choices = [
    disnake.OptionChoice(name="1 Minute", value="m1"),
    disnake.OptionChoice(name="5 Minutes", value="m5"),
    disnake.OptionChoice(name="10 Minutes", value="m10"),
    disnake.OptionChoice(name="15 Minutes", value="m15"),
    disnake.OptionChoice(name="30 Minutes", value="m30"),
    disnake.OptionChoice(name="1 Hour", value="m60"),
    disnake.OptionChoice(name="2 Hours", value="m120"),
    disnake.OptionChoice(name="4 Hours", value="m240"),
    disnake.OptionChoice(name="Daily", value="d1"),
    disnake.OptionChoice(name="Weekly", value="w")
]



@bot.command()
async def query(inter:disnake.AppCmdInter, ticker:str):
    await inter.send(view=QueryView(ticker))


@bot.command()
async def test(ctx, ticker):
    await ctx.send(view=QueryView(ticker))





bot.load_extensions('fudstop/cogs')
bot.run(os.environ.get('BOT'))


from list_sets.ticker_lists import most_active_tickers

from apis.webull.webull_trading import WebullTrading



class ReferenceMarket(WebullTrading):
    def __init__(self):
        self.webull_headers = self.headers





    



import asyncio
from fudstop.apis.polygonio.async_polygon_sdk import Polygon
from fudstop.list_sets.ticker_lists import most_active_tickers
polygon = Polygon()
async def get_2_4_8_rsi(ticker, timespan, window):
    windows_thresholds = {
        '2': (1, 99),  # (low_threshold, high_threshold)
        '4': (10, 90),
        '8': (20, 80),
    }
    try:

        # Assuming the polygon.rsi method returns an instance of RSI class
        rsi_obj = await polygon.rsi(ticker, timespan=timespan, window=str(window), limit=1)
        if rsi_obj and hasattr(rsi_obj, 'rsi_value'):
            rsi = rsi_obj.rsi_value
            if isinstance(rsi, list) and rsi:  # Ensure rsi is a list and not empty
                rsi_value = rsi[0]
                low_threshold, high_threshold = windows_thresholds[str(window)]
                if rsi_value <= low_threshold or rsi_value >= high_threshold:
                    return (ticker, rsi_value, window)
    except Exception as e:
        print(f"Error processing {ticker} for window {window}: {e}")
    return None

async def scan_2_4_8():
    windows = [2, 4, 8]
    tasks = [get_2_4_8_rsi(ticker, 'day', window) for ticker in most_active_tickers for window in windows]
    results = await asyncio.gather(*tasks)
    
    # Filter out None results
    valid_results = [result for result in results if result]
    
    # Now, check if any ticker meets the criteria for all windows
    ticker_results = {}
    for result in valid_results:
        ticker, rsi, window = result
        if ticker not in ticker_results:
            ticker_results[ticker] = []
        ticker_results[ticker].append(window)
    
    # Find tickers that have all windows
    all_results=[]
    for ticker, windows_list in ticker_results.items():
        if all(w in windows_list for w in windows):
            all_results.append(ticker)



    return all_results


import xml.etree.ElementTree as ET

import asyncio
#FORM_4

import asyncpg

# Modified function to recursively parse XML elements and prepare them for database insertion
import os

YOUR_API_KEY = os.environ.get("YOUR_POLYGON_KEY")

import re
import pandas as pd
from .polygonio.mapping import OPTIONS_EXCHANGES,option_condition_dict,STOCK_EXCHANGES,stock_condition_dict,TAPES

from colorsys import rgb_to_hsv
from bs4 import BeautifulSoup
import requests
from asyncio import Semaphore, TimeoutError
from typing import List, Union, Any, Tuple, Dict
from datetime import datetime
import pytz
import pandas as pd

# Function to format selected columns in a DataFrame
headers_sec = {'User-Agent': 'Fudstop https://discord.gg/fudstop', 'Content-Type': 'application/json'}

def format_value(value):
    return f"**{round(float(value), 2)}**" if value is not None else "**N/A**"

async def check_macd_sentiment(ticker, timespan, hist: list):
    if hist is not None:
        if hist is not None and len(hist) >= 3:
            
            last_three_values = hist[:3]
            if abs(last_three_values[0] - (-0.02)) < 0.04 and all(last_three_values[i] > last_three_values[i + 1] for i in range(len(last_three_values) - 1)):
                return 'bullish'

            if abs(last_three_values[0] - 0.02) < 0.04 and all(last_three_values[i] < last_three_values[i + 1] for i in range(len(last_three_values) - 1)):
                return 'bearish'
    else:
        return None
    
def flatten_object(obj, parent_key='', separator='_'):
    items = {}
    for k, v in obj.__dict__.items():
        new_key = f"{parent_key}{separator}{k}" if parent_key else k
        if isinstance(v, dict):
            items.update(flatten_object(v, new_key, separator=separator))
        elif hasattr(v, '__dict__'):  # Check if v is an object with attributes
            items.update(flatten_object(v, new_key, separator=separator))
        else:
            items[new_key] = v
    return items
def format_large_numbers_in_dataframe(df, columns_to_format):
    """
    Formats selected columns in a DataFrame to readable large numbers.
    """
    formatted_df = df.copy()
    for column in columns_to_format:
        if column in formatted_df.columns:
            formatted_df[column] = formatted_df[column].apply(format_large_number)
    return formatted_df
# Example function to format selected keys in a dictionary
def format_large_numbers_in_dict(data_dict, keys_to_format):
    """
    Formats selected keys in a dictionary to readable large numbers.
    """
    formatted_dict = data_dict.copy()
    for key in keys_to_format:
        if key in formatted_dict:
            formatted_dict[key] = format_large_number(formatted_dict[key])
    return formatted_dict
def format_large_number(num):
    if isinstance(num, list):
        return [format_large_number(n) for n in num]
    elif isinstance(num, str):
        try:
            num = float(num)
        except ValueError:
            return num  # or however you want to handle this case
    if num is None:
        return
    if abs(num) >= 1_000_000_000_000:
        return f"{num / 1_000_000_000_000:.2f} T"
    elif abs(num) >= 1_000_000_000:
        return f"{num / 1_000_000_000:.2f} B"
    elif abs(num) >= 1_000_000:
        return f"{num / 1_000_000:.2f} M"
    elif abs(num) >= 1_000:
        return f"{num / 1_000:.2f} K"
    else:
        return str(num)
    
def calculate_td9_series(df):
    setup_count = 0
    td9_series = pd.Series(index=df.index, dtype='Int64')  # Initialize a series with the same index as df

    for i in range(4, len(df)):
        if df['Close'][i] > df['Close'][i - 4]:
            setup_count += 1
        else:
            setup_count = 0

        if setup_count >= 9:
            td9_series.at[df.index[i]] = setup_count  # Store the count

    df['TD9'] = td9_series  # Add the series to the DataFrame

    return df

def calculate_setup(df):
    setup_count = 0
    for i in range(4, len(df)):
        if df['Close'][i] > df['Close'][i-4]:  # Assuming 'c' is the close price column
            setup_count += 1
        else:
            setup_count = 0
        
        if setup_count >= 9:
            return True
    return False
def calculate_countdown(df):
    countdown_count = 0
    for i in range(2, len(df)):
        if df['High'][i] > df['High'][i-2]:  # Assuming 'h' is the high price column
            countdown_count += 1
        else:
            countdown_count = 0
        
        if countdown_count >= 9:
            return True
    return False
def send_to_discord(image_path, webhook_url):
    with open(image_path, 'rb') as f:
        files = {'file': ('image.png', f)}
        payload = {
            "embeds": [
                {
                    "title": "TD9 Chart",
                    "image": {
                        "url": "attachment://image.png"
                    }
                }
            ]
        }
        headers = {'Content-Type': 'multipart/form-data', 'Authorization': f"Bearer {os.environ.get('YOUR_DISCORD_HTTP_TOKEN')}"}
        r = requests.post(webhook_url, headers=headers, files=files, json=payload)
        if r.status_code == 204:
            print("Successfully sent image to Discord.")
        else:
            print(f"Failed to send image to Discord. Status Code: {r.status_code}")

def human_readable(string):
    try:
        match = re.search(r'(\w{1,5})(\d{2})(\d{2})(\d{2})([CP])(\d+)', string) #looks for the options symbol in O: format
        underlying_symbol, year, month, day, call_put, strike_price = match.groups()
    except TypeError:
        underlying_symbol = f"AMC"
        year = "23"
        month = "02"
        day = "17"
        call_put = "CALL"
        strike_price = "380000"
    
    expiry_date = month + '/' + day + '/' + '20' + year
    if call_put == 'C':
        call_put = 'Call'
    else:
        call_put = 'Put'
    strike_price = '${:.2f}'.format(float(strike_price)/1000)
    return "{} {} {} Expiring {}".format(underlying_symbol, strike_price, call_put, expiry_date)

@staticmethod
def get_human_readable_string(string):
    result = {}
    try:
        match = re.search(r'(\w{1,5})(\d{2})(\d{2})(\d{2})([CP])(\d+)', string)
        underlying_symbol, year, month, day, call_put, strike_price = match.groups()
    except TypeError:
        underlying_symbol = "AMC"
        year = "23"
        month = "02"
        day = "17"
        call_put = "CALL"
        strike_price = "380000"

    expiry_date = '20' + year + '-' + month + '-' + day
    call_put = 'Call' if call_put == 'C' else 'Put'
    strike_price = float(strike_price) / 1000
    result['underlying_symbol'] = underlying_symbol
    result['strike_price'] = strike_price
    result['call_put'] = call_put
    result['expiry_date'] = expiry_date
    return result

def flatten(item, parent_key='', separator='_'):
    items = {}
    if isinstance(item, dict):
        for k, v in item.items():
            new_key = f"{parent_key}{separator}{k}" if parent_key else k
            if isinstance(v, dict):
                items.update(flatten(v, new_key, separator=separator))
            elif isinstance(v, list):
                for i, elem in enumerate(v):
                    items.update(flatten(elem, f"{new_key}_{i}", separator=separator))
            else:
                items[new_key] = v
    elif isinstance(item, list):
        for i, elem in enumerate(item):
            items.update(flatten(elem, f"{parent_key}_{i}", separator=separator))
    else:
        items[parent_key] = item
    return items

def flatten_list_of_dicts(lst: List[Union[Dict, List]]) -> List[Dict]:
    return [flatten(item) for item in lst]




def is_current_candle_td9(df: pd.DataFrame) -> bool:
    """
    Check if the latest (current) candle in the DataFrame is a TD9.
    """
    # Ensure we have enough data for the check.
    if df.shape[0] < 10:  # We need 10 candles to compare the 9th with the 0th
        return False

    # Get the first 10 candles (assuming the DataFrame is sorted in descending order by time).
    first_ten = df.head(10).reset_index(drop=True)
    
    # Check for a Buy Setup TD9
    buy_setup = first_ten.loc[0, 'Close'] < first_ten.loc[9, 'Close']
    
    # Check for a Sell Setup TD9
    sell_setup = first_ten.loc[0, 'Close'] > first_ten.loc[9, 'Close']

    return buy_setup or sell_setup
# ... [your code] ...

# Dynam


async def check_macd_sentiment(ticker, timespan, hist: list):
    if hist is not None:
        if hist is not None and len(hist) >= 3:
            
            last_three_values = hist[:3]
            if abs(last_three_values[0] - (-0.02)) < 0.04 and all(last_three_values[i] > last_three_values[i + 1] for i in range(len(last_three_values) - 1)):
                return 'bullish'

            if abs(last_three_values[0] - 0.02) < 0.04 and all(last_three_values[i] < last_three_values[i + 1] for i in range(len(last_three_values) - 1)):
                return 'bearish'
    else:
        return None
    
def convert_to_ns_datetime(unix_timestamp_str):
    # Convert Unix timestamp string to integer and then to seconds
    unix_timestamp = int(unix_timestamp_str) / 1000.0
    
    # Convert to datetime object in UTC
    dt_utc = datetime.utcfromtimestamp(unix_timestamp)
    
    # Localize to UTC
    dt_utc = pytz.utc.localize(dt_utc)
    
    # Convert to Eastern Time
    dt_et = dt_utc.astimezone(pytz.timezone('US/Eastern'))
    
    # Remove the timezone offset information, if you want to
    dt_et = dt_et.replace(tzinfo=None)
    
    return dt_et   
def convert_to_datetime(unix_timestamp_str):
    # Convert Unix timestamp string to integer
    unix_timestamp = int(unix_timestamp_str)
    
    # Convert to datetime object in UTC
    dt_utc = datetime.utcfromtimestamp(unix_timestamp)
    
    # Localize to UTC
    dt_utc = pytz.utc.localize(dt_utc)
    
    # Convert to Eastern Time
    dt_et = dt_utc.astimezone(pytz.timezone('US/Eastern'))
    
    # Remove the timezone offset information, if you want to
    dt_et = dt_et.replace(tzinfo=None)
    

    return dt_et

def convert_to_datetime_or_str(input_str):
    try:
        # If it's a Unix timestamp, convert it
        unix_timestamp = int(input_str)
        dt_utc = datetime.utcfromtimestamp(unix_timestamp)
        dt_utc = pytz.utc.localize(dt_utc)
        dt_et = dt_utc.astimezone(pytz.timezone('US/Eastern'))
        return dt_et.replace(tzinfo=None)
    except ValueError:
        # If it's a date string, parse it
        return datetime.strptime(input_str, '%B %d, %Y')

def convert_datetime_list(timestamps, unit='ms'):
    """
    Convert a list of Unix timestamps to datetime objects.

    Parameters:
    - timestamps: list of Unix timestamps
    - unit: the unit of the timestamp (default is 's' for seconds)

    Returns:
    - list of datetime objects
    """
    dt_series = pd.Series(pd.to_datetime(timestamps, unit=unit, utc=True))
    dt_series = dt_series.dt.tz_localize(None)
    return dt_series.tolist()

# Function to convert timestamps to Eastern Time (ET)
def convert_to_et(timestamp):
    # Assuming timestamps are in UTC, you can convert them to ET
    utc_time = datetime.strptime(timestamp, '%Y-%m-%dT%H:%M:%S.%fZ')
    et_timezone = pytz.timezone('US/Eastern')  # Eastern Time (ET)
    et_time = utc_time.astimezone(et_timezone)
    return et_time.strftime('%Y-%m-%d %H:%M:%S %Z')
def calculate_days_to_expiry(expiry_str, timestamp):
    expiry = datetime.strptime(expiry_str, '%m/%d/%Y').date()
    return (expiry - timestamp.date()).days
def calculate_price_to_strike(price, strike):
    return price / strike if strike != 0 else 0


def map_conditions(conditions):
    if isinstance(conditions, list):
        return [option_condition_dict.get(cond, cond) for cond in conditions]
    else:
        return conditions
def map_stock_conditions(conditions):
    if isinstance(conditions, list):
        return [stock_condition_dict.get(cond, cond) for cond in conditions]
    else:
        return conditions
def count_itm_otm(group):
    underlying_price = group['underlying_price'].iloc[0]  # Assuming the underlying price is the same within each expiry group
    itm_call = len(group[(group['call_put'] == 'call') & (group['strike'] < underlying_price)])
    otm_call = len(group[(group['call_put'] == 'call') & (group['strike'] >= underlying_price)])
    itm_put = len(group[(group['call_put'] == 'put') & (group['strike'] > underlying_price)])
    otm_put = len(group[(group['call_put'] == 'put') & (group['strike'] <= underlying_price)])

    return pd.Series({
        'ITM_calls': itm_call,
        'OTM_calls': otm_call,
        'ITM_puts': itm_put,
        'OTM_puts': otm_put
    })
def calculate_candlestick(data, interval):
    open_price = data[0]['open_price']
    close_price = data[-1]['close_price']
    high_price = max(item['high_price'] for item in data)
    low_price = min(item['low_price'] for item in data)
    volume = sum(item['volume'] for item in data)

    return {
        'open_price': open_price,
        'close_price': close_price,
        'high_price': high_price,
        'low_price': low_price,
        'volume': volume
    }

def to_unix_timestamp_eastern(timestamp_ns):
    timestamp_eastern = to_datetime_eastern(timestamp_ns)
    return int(timestamp_eastern.timestamp())
def to_datetime_eastern(timestamp_ns):
    # Convert the timestamp to a pandas datetime object in UTC
    timestamp_utc = pd.to_datetime(timestamp_ns, unit='ns').tz_localize('UTC')

    # Convert the timestamp to the US Eastern timezone
    timestamp_eastern = timestamp_utc.tz_convert('US/Eastern')

    return timestamp_eastern



# Function to traverse the XML tree and collect unique tags and keys
def traverse_tree(element, unique_tags, unique_keys):
    unique_tags.add(element.tag)
    for key in element.keys():
        unique_keys.add(key)
    for child in element:
        traverse_tree(child, unique_tags, unique_keys)

# Function to traverse the XML tree and extract information
def traverse_and_extract(element, target_tags, extracted_data):
    if element.tag in target_tags:
        extracted_data[element.tag] = element.text
    for child in element:
        traverse_and_extract(child, target_tags, extracted_data)
# Function to recursively parse XML elements and return them in a dictionary format
def parse_element(element, parsed=None):
    if parsed is None:
        parsed = {}
    
    for child in element:
        # Skip elements without values
        if child.text is None or child.text.strip() == '':
            continue

        # Handle duplicate tags by converting them into lists
        if child.tag in parsed:
            if not isinstance(parsed[child.tag], list):
                parsed[child.tag] = [parsed[child.tag]]
            parsed[child.tag].append(parse_element(child, {}))
        else:
            parsed[child.tag] = parse_element(child, {})
        
        # Store the element value if it has one
        if child.text.strip():
            parsed[child.tag]['value'] = child.text.strip()
            
    return parsed

def download_xml_file(url, file_path):
    response = requests.get(url, headers=headers_sec)
    if response.status_code == 200:
        with open(file_path, 'wb') as f:
            f.write(response.content)
        print(f"File downloaded successfully at {file_path}")
    else:
        print(f"Failed to download the file. Status code: {response.status_code}")



def prepare_data_for_insertion(element, parsed=None, current_table=None, current_record=None):
    if parsed is None:
        parsed = {'ownershipDocument': [], 'issuer': [], 'reportingOwner': [], 
                  'reportingOwnerAddress': [], 'nonDerivativeTransaction': [], 
                  'transactionAmounts': [], 'postTransactionAmounts': [], 'footnote': []}
    
    if current_table and current_record is not None:
        # Add the element value to the current table's last record
        if element.text and element.text.strip():
            current_record[element.tag] = element.text.strip()

    # Determine the current table based on the element tag
    if element.tag in parsed:
        current_table = element.tag
        new_record = {}
        parsed[current_table].append(new_record)
        current_record = new_record

    for child in element:
        # Recursive call to handle nested elements
        prepare_data_for_insertion(child, parsed, current_table, current_record)
            
    return parsed

def safe_divide(a, b):
    if a is None or b is None or b == 0:
        return None
    return a / b

def safe_subtract(a, b):
    if a is None or b is None:
        return None
    return a - b

def safe_multiply(a, b):
    if a is None or b is None:
        return None
    return a * b

def safe_max(a, b):
    if a is None and b is None:
        return None
    if a is None:
        return b
    if b is None:
        return a
    return max(a, b)



# Function to get image URL from a webpage
def get_first_image_url(webpage_url):
    # Download the webpage
    response = requests.get(webpage_url)
    if response.status_code != 200:
        return None  # Failed to download

    # Parse the HTML
    soup = BeautifulSoup(response.text, 'html.parser')

    # Find the first image tag
    img_tag = soup.find('img')
    
    if img_tag is None:
        return None  # No image found

    # Extract the image URL
    img_url = img_tag.get('src')

    return img_url
def parse_to_dataframe(data):
    # Extracting relevant fields from the JSON response
    results = data['results'] if 'results' in data else data
    parsed_data = []
    if results is not data:
        for result in results:
            parsed_data.append({
                "article_url": result.get("article_url", None),
                "author": result.get("author", None),
                "description": result.get("description", None),
                "id": result.get("id", None),
                "published_utc": result.get("published_utc", None),
                "publisher_name": result.get("publisher", {}).get("name", None),
                "tickers": ", ".join(result.get("tickers", [])),
                "title": result.get("title", None)
            })
        
        # Create DataFrame
        df = pd.DataFrame(parsed_data)
        return df
    
    # Initialize an empty list to store flattened dictionaries
    flattened_data = []

    # Iterate through the data
    for item in data:
        flat_item = {}

        # Recursively flatten nested dictionaries
        def flatten_dict(d, parent_key=''):
            for key, value in d.items():
                new_key = parent_key + '_' + key if parent_key else key
                if isinstance(value, dict):
                    flatten_dict(value, new_key)
                else:
                    # Convert list values to strings
                    if isinstance(value, list):
                        value = str(value)
                    flat_item[new_key] = value

        flatten_dict(item)

        # Append the flattened dictionary to the list
        flattened_data.append(flat_item)

    # Create a DataFrame from the flattened data
    df = pd.DataFrame(flattened_data)

    # Display the DataFrame
    print(df)
        

def get_first_index_from_dict(data_dict):
    first_index_data_dict = {}
    for key, value in data_dict.items():
        if value:  # Check if the value (which should be a list) is not empty
            first_index_data_dict[key] = value[0]
        else:
            first_index_data_dict[key] = None  # or some default value
    return first_index_data_dict


def calculate_percent_decrease(open_price, close_price):
    percent_decrease = ((open_price - close_price) / close_price) * 100
    return percent_decrease


import aiohttp
import asyncio

from urllib.parse import urlencode
# Fetch all URLs

async def fetch_url(session, url):
    async with session.get(url) as response:
        if response.status == 200:
            data = await response.json()
            return data
        else:
            print(f"Error: {response.status}")
            return None




async def process_contract(data, db):
        


    if data is not None:
        for contract in data:
            try:

                option_symbol = contract.get('ticker')
                expiry = contract.get('details').get('expiration_date')
                strike = contract.get('details').get('strike_price')
                call_put = contract.get('details').get('contract_type')
                underlying_symbol = contract.get('underlying_asset').get('ticker')
                underlying_price = contract.get('underlying_asset').get('price')
                change_to_breakeven = contract.get('underlying_asset').get('change_to_breakeven')

                

                trade_size = contract.get('last_trade').get('size')
                trade_price = contract.get('last_trade').get('price')
                trade_conditions = contract.get('last_trade', None).get('conditions', None)
                sip_timestamp = contract.get('last_trade', None).get('sip_timestamp', None)
                if sip_timestamp is not None:
                    sip_timestamp = datetime.fromtimestamp(sip_timestamp / 1e9)
                trade_conditions = [option_condition_dict.get(c) for c in trade_conditions] if trade_conditions is not None else []
                trade_conditions = trade_conditions[0]
                trade_exchange = OPTIONS_EXCHANGES.get(contract.get('last_trade').get('exchange', None))
                break_even_price = contract.get('break_even_price', None)
                implied_volatility = contract.get('implied_volatility', None)
                open_interest = contract.get('open_interest', None)
                name = contract.get('name', None)
                open = contract.get('session').get('open', None)
                high = contract.get('session').get('high', None)
                low = contract.get('session').get('low', None)
                close = contract.get('session').get('close', None)
                volume = contract.get('session').get('volume', None)
                prev_close = contract.get('session').get('previous_close')
                early_trading_change_percent = contract.get('session').get('early_trading_change_percent')
                change = contract.get('session').get('change')
                early_trading_change = contract.get('session').get('early_trading_change')
                change_percent = contract.get('session').get('change_percent', None)

                delta = contract.get('greeks').get('delta')
                vega = contract.get('greeks').get('vega')
                theta = contract.get('greeks').get('theta')
                gamma = contract.get('greeks').get('gamma')

                ask = contract.get('last_quote').get('ask')
                bid = contract.get('last_quote').get('bid')
                ask_size = contract.get('last_quote').get('ask_size')
                bid_size = contract.get('last_quote').get('bid_size')
                ask_exchange = OPTIONS_EXCHANGES.get(contract.get('last_quote').get('ask_exchange'))
                bid_exchange = OPTIONS_EXCHANGES.get(contract.get('last_quote').get('bid_exchange'))


                data_dict = {
                    'option_symbol': option_symbol,
                    'expiry': expiry,
                    'strike': strike,
                    'call_put': call_put,
                    'underlying_symbol': underlying_symbol,
                    'underlying_price': underlying_price,
                    'change_to_breakeven': change_to_breakeven,
                    'trade_size': trade_size,
                    'trade_price': trade_price,
                    'trade_conditions': trade_conditions,
                    'sip_timestamp': sip_timestamp,
                    'trade_exchange': trade_exchange,
                    'break_even_price': break_even_price,
                    'implied_volatility': implied_volatility,
                    'open_interest': open_interest,
                    'name': name,
                    'open': open,
                    'high': high,
                    'low': low,
                    'close': close,
                    'volume': volume,
                    'prev_close': prev_close,
                    'early_trading_change_percent': early_trading_change_percent,
                    'change': change,
                    'early_trading_change': early_trading_change,
                    'change_percent': change_percent,
                    'delta': delta,
                    'vega': vega,
                    'theta': theta,
                    'gamma': gamma,
                    'ask': ask,
                    'bid': bid,
                    'ask_size': ask_size,
                    'bid_size': bid_size,
                    'ask_exchange': ask_exchange,
                    'bid_exchange': bid_exchange
                }


                return data_dict
            except Exception as e:
                # Handle any other exceptions that might occur
                print(f"An error occurred: {e}")
from asyncio import Semaphore
sema = Semaphore(5)


async def fetch_page(url):
    try:
        async with aiohttp.ClientSession() as session, session.get(url) as response:
            response.raise_for_status()
            return await response.json()
    except TimeoutError:
        print(f"Timeout when accessing {url}")
    except aiohttp.ClientResponseError as http_err:
        print(f"HTTP error occurred: {http_err}")
    except Exception as err:
        print(f"An error occurred: {err}")

async def paginate_concurrent(url, as_dataframe=False, concurrency=25):
    all_results = []

    
    async with aiohttp.ClientSession() as session:
        pages_to_fetch = [url]
        
        while pages_to_fetch:
            tasks = []
            
            for _ in range(min(concurrency, len(pages_to_fetch))):
                next_url = pages_to_fetch.pop(0)
                tasks.append(fetch_page(next_url))
                
            results = await asyncio.gather(*tasks)
            if results is not None:
                for data in results:
                    if data is not None:
                        if "results" in data:
                            all_results.extend(data["results"])
                            
                        next_url = data.get("next_url")
                        if next_url:
                            next_url += f'&{urlencode({"apiKey": "iWivQU_dAR5CZHKpd15bEBApfWfBZgJ5"})}'
                            pages_to_fetch.append(next_url)
                    else:
                        break
    if as_dataframe:
        import pandas as pd
        return pd.DataFrame(all_results)
    else:
        return all_results
    
def describe_color(rgb_tuple):
    red, green, blue = rgb_tuple
    # Normalize the RGB values
    red, green, blue = red / 255.0, green / 255.0, blue / 255.0

    # Convert RGB to HSV
    hue, saturation, value = rgb_to_hsv(red, green, blue)
    
    # Determine the color based on hue
    hue_degree = hue * 360
    if saturation < 0.1 and value > 0.9:
        return "White"
    elif saturation < 0.2 and value < 0.2:
        return "Black"
    elif saturation < 0.2:
        return "Gray"
    elif hue_degree >= 0 and hue_degree < 12:
        return "Red"
    elif hue_degree >= 12 and hue_degree < 35:
        return "Orange"
    elif hue_degree >= 35 and hue_degree < 85:
        return "Yellow"
    elif hue_degree >= 85 and hue_degree < 170:
        return "Green"
    elif hue_degree >= 170 and hue_degree < 260:
        return "Blue"
    elif hue_degree >= 260 and hue_degree < 320:
        return "Purple"
    else:
        return "Red"
def parse_operation(operation):
    # Initialize an empty dictionary to store the parsed data
    parsed_data = {}

    # Extract the main attributes from the operation dictionary
    main_attrs = [
        'operationId', 'auctionStatus', 'operationDate', 'settlementDate', 
        'maturityDate', 'operationType', 'operationMethod', 'settlementType',
        'termCalenderDays', 'term', 'releaseTime', 'closeTime', 'note',
        'lastUpdated', 'participatingCpty', 'acceptedCpty', 'totalAmtSubmitted',
        'totalAmtAccepted'
    ]
    for attr in main_attrs:
        parsed_data[attr] = operation.get(attr)

    # Check if the 'details' key exists and is a non-empty list
    details = operation.get('details')
    if details and isinstance(details, list) and len(details) > 0:
        # Assuming details is a list and you're working with the first item
        details_dict = details[0]
        # Extract the attributes from the details dictionary
        detail_attrs = [
            'securityType', 'amtSubmitted', 'amtAccepted',
            'percentOfferingRate', 'percentAwardRate'
        ]
        for attr in detail_attrs:
            parsed_data[attr] = details_dict.get(attr)

    return parsed_data    
# Extend the existing function
def decimal_to_color(decimal_color):
    # Convert the decimal number to hexadecimal
    hex_color = hex(decimal_color)[2:].zfill(6).upper()
    
    # Extract the RGB components
    red = int(hex_color[0:2], 16)
    green = int(hex_color[2:4], 16)
    blue = int(hex_color[4:6], 16)
    
    # Describe the color
    color_name = describe_color((red, green, blue))
    
    return color_name

async def paginate_tickers(url, as_dataframe=False, concurrency=5):
    all_results = []

    
    async with aiohttp.ClientSession() as session:
        pages_to_fetch = [url]
        
        while pages_to_fetch:
            tasks = []
            
            for _ in range(min(concurrency, len(pages_to_fetch))):
                next_url = pages_to_fetch.pop(0)
                tasks.append(fetch_page(next_url))
                
            results = await asyncio.gather(*tasks)
            if results is not None:
                for data in results:
                    if data is not None:
                        if "tickers" in data:
                            all_results.extend(data["tickers"])
                            
                        next_url = data.get("next_url")
                        if next_url:
                            next_url += f'&{urlencode({"apiKey": YOUR_API_KEY})}'
                            pages_to_fetch.append(next_url)
                    
    if as_dataframe:
        import pandas as pd
        return pd.DataFrame(all_results)
    else:
        return all_results









async def fetch_and_parse_data(url):
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            response.raise_for_status()
            data = await response.json()
            results = data.get('results', [])
            
            # Flatten nested dictionaries
            flattened_results = [flatten_dict(result) for result in results]
            
            return flattened_results 

def flatten_dict(d, parent_key='', sep='.'):
    items = {}
    for k, v in d.items():
        new_key = f"{parent_key}{sep}{k}" if parent_key else k
        if isinstance(v, dict):
            items.update(flatten_dict(v, new_key, sep=sep))
        else:
            items[new_key] = v
    return items
def rename_keys(original_dict, key_mapping):
    return {key_mapping.get(k, k): v for k, v in original_dict.items()}

async def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

selected_ticker = None

import time
def last_unix_interval(threshold) -> int:
    now = datetime.now()
    minute = now.minute

    # Round down to the nearest 30-minute mark
    if minute >= threshold:
        rounded_minute = 30
    else:
        rounded_minute = 0

    # Construct the datetime object for the last 30-minute interval
    last_30_min_interval = now.replace(minute=rounded_minute, second=0, microsecond=0)

    # Convert to Unix timestamp
    unix_timestamp = int(time.mktime(last_30_min_interval.timetuple()))

    return unix_timestamp



def remove_html_tags(text):
    clean = re.compile('<.*?>')
    return re.sub(clean, '', text)
def convert_str_to_datetime(date_time_str):
    # Parsing time and date from the string
    date_time_str = str(date_time_str)
    time_str, am_pm, _, _, month, day_with_comma, year = date_time_str.split()

    # Debug print for inspection
    print(f"Debug: {date_time_str}")

    hour, minute = map(int, time_str.split(':'))

    # Adjusting for AM/PM
    if am_pm == 'PM' and hour != 12:
        hour += 12
    elif am_pm == 'AM' and hour == 12:
        hour = 0

    # Removing comma from the day
    day = int(day_with_comma.replace(",", ""))

    # Constructing datetime object with the US/Eastern timezone
    dt_string = f"{year}-{month}-{day:02} {hour:02}:{minute:02}"
    dt_et = datetime.strptime(dt_string, '%Y-%B-%d %H:%M')
    eastern = pytz.timezone('US/Eastern')
    dt_et = eastern.localize(dt_et)

    # Convert to desired timezone or manipulate further if necessary
    return dt_et
def map_months(month_str):
    month_dict = {
        "January": "01",
        "February": "02",
        "March": "03",
        "April": "04",
        "May": "05",
        "June": "06",
        "July": "07",
        "August": "08",
        "September": "09",
        "October": "10",
        "November": "11",
        "December": "12"
    }
    return month_dict.get(month_str, "Invalid month")



def current_time_to_unix() -> int:
    # Get current time
    now = datetime.now()
    
    # Convert to Unix timestamp
    unix_timestamp = int(time.mktime(now.timetuple()))
    
    return unix_timestamp
    
def convert_timestamp_to_human_readable(url: str) -> str:
    try:
        timestamp_str = url.split("timestamp=")[-1]
        timestamp = int(timestamp_str)
    except (ValueError, IndexError):
        return "Invalid URL or timestamp"

    dt_object = datetime.fromtimestamp(timestamp)
    human_readable_date = dt_object.strftime('%Y-%m-%d %H:%M:%S')

    return human_readable_date

def convert_to_yymmdd(expiry_str):
    expiry_date = datetime.strptime(expiry_str, '%Y-%m-%d')
    return expiry_date.strftime('%y%m%d')


def parse_element(element, parent_key='', parsed_dict={}):
    children = list(element)
    if parent_key:
        parent_key += '.'
    
    if children:
        for child in children:
            parse_element(child, parent_key + child.tag, parsed_dict)
    else:
        parsed_dict[parent_key[:-1]] = element.text
def shorten_form4_keys(data_dict):
    shortened_dict = {}
    for key, value in data_dict.items():
        new_key = key
        # Remove common prefixes for 'issuer'
        new_key = new_key.replace('issuer_', '')
        
        # Remove common prefixes for 'reportingOwner'
        new_key = new_key.replace('reportingOwner_', '')
        new_key = new_key.replace('reportingOwnerId_', '')
        new_key = new_key.replace('reportingOwnerAddress_', '')
        new_key = new_key.replace('reportingOwnerRelationship_', '')
        
        # Remove common prefixes for 'nonDerivativeTable' and 'nonDerivativeTransaction'
        new_key = new_key.replace('nonDerivativeTable_', '')
        new_key = new_key.replace('nonDerivativeTransaction_', '')
        
        # Remove common prefixes for 'transactionCoding', 'transactionAmounts', 'postTransactionAmounts', and 'ownershipNature'
        new_key = new_key.replace('transactionCoding_', '')
        new_key = new_key.replace('transactionAmounts_', '')
        new_key = new_key.replace('postTransactionAmounts_', '')
        new_key = new_key.replace('ownershipNature_', '')
        
        # Remove common prefixes for 'securityTitle', 'transactionDate', 'transactionTimeliness', etc.
        new_key = new_key.replace('securityTitle_', '')
        new_key = new_key.replace('transactionDate_', '')
        new_key = new_key.replace('transactionTimeliness_', '')
        
        # Remove common prefixes for 'footnotes' and 'ownerSignature'
        new_key = new_key.replace('footnotes_', '')
        new_key = new_key.replace('ownerSignature_', '')
        
        # Add more replacements as needed
        
        shortened_dict[new_key] = value
    return shortened_dict


# Function to recursively find all fields and values in an XML ElementTree
def extract_fields_recursive(element, parent_key='', results={}):
    for child in element:
        key = f"{parent_key}.{child.tag}" if parent_key else child.tag
        if len(child) > 0:
            extract_fields_recursive(child, key, results)
        else:
            results[key] = child.text
    return results
import matplotlib.pyplot as plt
def save_df_as_image(df, image_path):
    fig, ax = plt.subplots(figsize=(8, 4))
    ax.axis('tight')
    ax.axis('off')
    plt.table(cellText=df.values, colLabels=df.columns, cellLoc='center', loc='center')
    plt.savefig(image_path)
    plt.close(fig)


import natsort
from PIL import Image

def autocrop_image(image: Any, border=0) -> Any:
    """Crop empty space from PIL image

    Parameters
    ----------
    image : Image
        PIL image to crop
    border : int, optional
        scale border outwards, by default 0

    Returns
    -------
    Image
        Cropped image
    """
    bbox = image.getbbox()
    image = image.crop(bbox)
    (width, height) = image.size
    width += border * 2
    height += border * 2
    cropped_image = Image.new("RGBA", (width, height), (0, 0, 0, 0))
    cropped_image.paste(image, (border, border))
    return cropped_image


conversion_mapping = {
    "K": 1_000,
    "M": 1_000_000,
}



# Approach 2: Fill Missing Values
def fill_missing_values(data_dict):
    max_length = max(len(v) for v in data_dict.values())
    for k, v in data_dict.items():
        if len(v) < max_length:
            data_dict[k] = v + [None] * (max_length - len(v))



all_units = "|".join(conversion_mapping.keys())
float_re = natsort.numeric_regex_chooser(natsort.ns.FLOAT | natsort.ns.SIGNED)
unit_finder = re.compile(rf"({float_re})\s*({all_units})", re.IGNORECASE)
from kaleido.scopes.plotly import PlotlyScope

scope = PlotlyScope()
import io
import plotly.graph_objects as go
import uuid
def save_image(filename: str, fig: go.Figure = None, bytesIO: io.BytesIO = None) -> str:
    """Takes go.Figure or io.BytesIO object, adds uuid to filename, autocrops, and saves

    Parameters
    ----------
    filename : str
        Name to save image as
    fig : go.Figure, optional
        Table object to autocrop and save, by default None
    bytesIO : io.BytesIO, optional
        BystesIO object to autocrop and save, by default None

    Returns
    -------
    str
        filename with UUID added to use for bot processing

    Raises
    ------
    Exception
        Function requires a go.Figure or BytesIO object
    """
    imagefile = "image.jpg"

    if fig:
        # Transform Fig into PNG with Running Scope. Returns image bytes
        fig = scope.transform(fig, scale=3, format="png")
        imgbytes = io.BytesIO(fig)
    elif bytesIO:
        imgbytes = bytesIO
    else:
        raise Exception("Function requires a go.Figure or io.BytesIO object")

    image = Image.open(imgbytes)
    image = autocrop_image(image, 0)
    imgbytes.seek(0)
    image.save(imagefile, "jpg", quality=100)
    image.close()

    return imagefile



import pandas as pd

class FinancialStatement:
    def __init__(self, r):
        data = r['data']
        self.quote_id = [i.get("quoteId", None) for i in data]
        self.type = [i.get("type", None) for i in data]
        self.fiscal_year = [i.get("fiscalYear", None) for i in data]
        self.fiscal_period = [i.get("fiscalPeriod", None) for i in data]
        self.end_date = [i.get("endDate", None) for i in data]
        self.currency_id = [i.get("currencyId", None) for i in data]
        self.publish_date = [i.get("publishDate", None) for i in data]
        self.total_revenue = [i.get("totalRevenue", None) for i in data]
        self.revenue = [i.get("revenue", None) for i in data]
        self.cost_of_revenue_total = [i.get("costofRevenueTotal", None) for i in data]
        self.gross_profit = [i.get("grossProfit", None) for i in data]
        self.operating_expense = [i.get("operatingExpense", None) for i in data]
        self.sell_gen_admin_expenses = [i.get("sellGenAdminExpenses", None) for i in data]
        self.depreciation_and_amortization = [i.get("depreciationAndAmortization", None) for i in data]
        self.inter_expse_inc_net_oper = [i.get("interExpseIncNetOper", None) for i in data]
        self.unusual_expense_income = [i.get("unusualExpenseIncome", None) for i in data]
        self.operating_income = [i.get("operatingIncome", None) for i in data]
        self.inter_inc_expse_net_non_oper = [i.get("interIncExpseNetNonOper", None) for i in data]
        self.net_income_before_tax = [i.get("netIncomeBeforeTax", None) for i in data]
        self.income_tax = [i.get("incomeTax", None) for i in data]
        self.net_income_after_tax = [i.get("netIncomeAfterTax", None) for i in data]
        self.net_income_before_extra = [i.get("netIncomeBeforeExtra", None) for i in data]
        self.total_extraordinary_items = [i.get("totalExtraordinaryItems", None) for i in data]
        self.net_income = [i.get("netIncome", None) for i in data]
        self.income_avaito_com_excl_extra_ord = [i.get("incomeAvaitoComExclExtraOrd", None) for i in data]
        self.income_avaito_com_incl_extra_ord = [i.get("incomeAvaitoComInclExtraOrd", None) for i in data]
        self.diluted_net_income = [i.get("dilutedNetIncome", None) for i in data]
        self.diluted_weighted_average_shares = [i.get("dilutedWeightedAverageShares", None) for i in data]
        self.diluted_eps_excl_extra_items = [i.get("dilutedEPSExclExtraItems", None) for i in data]
        self.diluted_eps_incl_extra_items = [i.get("dilutedEPSInclExtraItems", None) for i in data]
        self.diluted_normalized_eps = [i.get("dilutedNormalizedEPS", None) for i in data]
        self.operating_profit = [i.get("operatingProfit", None) for i in data]
        self.earning_after_tax = [i.get("earningAfterTax", None) for i in data]
        self.earning_before_tax = [i.get("earningBeforeTax", None) for i in data]
        
        self.data_dict ={
            'quoteId': self.quote_id,
            'type': self.type,
            'fiscalYear': self.fiscal_year,
            'fiscalPeriod': self.fiscal_period,
            'endDate': self.end_date,
            'currencyId': self.currency_id,
            'publishDate': self.publish_date,
            'totalRevenue': self.total_revenue,
            'revenue': self.revenue,
            'costofRevenueTotal': self.cost_of_revenue_total,
            'grossProfit': self.gross_profit,
            'operatingExpense': self.operating_expense,
            'sellGenAdminExpenses': self.sell_gen_admin_expenses,
            'depreciationAndAmortization': self.depreciation_and_amortization,
            'interExpseIncNetOper': self.inter_expse_inc_net_oper,
            'unusualExpenseIncome': self.unusual_expense_income,
            'operatingIncome': self.operating_income,
            'interIncExpseNetNonOper': self.inter_inc_expse_net_non_oper,
            'netIncomeBeforeTax': self.net_income_before_tax,
            'incomeTax': self.income_tax,
            'netIncomeAfterTax': self.net_income_after_tax,
            'netIncomeBeforeExtra': self.net_income_before_extra,
            'totalExtraordinaryItems': self.total_extraordinary_items,
            'netIncome': self.net_income,
            'incomeAvaitoComExclExtraOrd': self.income_avaito_com_excl_extra_ord,
            'incomeAvaitoComInclExtraOrd': self.income_avaito_com_incl_extra_ord,
            'dilutedNetIncome': self.diluted_net_income,
            'dilutedWeightedAverageShares': self.diluted_weighted_average_shares,
            'dilutedEPSExclExtraItems': self.diluted_eps_excl_extra_items,
            'dilutedEPSInclExtraItems': self.diluted_eps_incl_extra_items,
            'dilutedNormalizedEPS': self.diluted_normalized_eps,
            'operatingProfit': self.operating_profit,
            'earningAfterTax': self.earning_after_tax,
            'earningBeforeTax': self.earning_before_tax
        }

        self.df = pd.DataFrame(self.data_dict)
    @classmethod
    def from_dict(cls, data):
        return cls(data)

    def __repr__(self):
        return f'<FinancialStatement quote_id={self.quote_id} fiscal_year={self.fiscal_year} fiscal_period={self.fiscal_period}>'


class CashFlow:
    def __init__(self, r):
        data = r['data']
        self.quoteid = [i.get('quoteId', None) for i in data]
        self.type = [i.get('type', None) for i in data]
        self.fiscal_year = [i.get('fiscalYear', None) for i in data]
        self.fiscal_period = [i.get('fiscalPeriod', None) for i in data]
        self.end_date = [i.get('endDate', None) for i in data]
        self.currency_id = [i.get('currencyId', None) for i in data]
        self.publish_date = [i.get('publishDate', None) for i in data]
        self.cash_from_operating_activities = [i.get('cashfromOperatingActivities', None) for i in data]
        self.net_income = [i.get('netIncome', None) for i in data]
        self.depreciation_and_amortization = [i.get('depreciationAndAmortization', None) for i in data]
        self.deferred_taxes = [i.get('deferredTaxes', None) for i in data]
        self.non_cash_items = [i.get('nonCashItems', None) for i in data]
        self.changes_in_working_capital = [i.get('changesinWorkingCapital', None) for i in data]
        self.cash_from_investing_activities = [i.get('cashfromInvestingActivities', None) for i in data]
        self.capital_expenditures = [i.get('capitalExpenditures', None) for i in data]
        self.other_investing_cashflow_items_total = [i.get('otherInvestingCashFlowItemsTotal', None) for i in data]
        self.cash_from_financing_activities = [i.get('cashfromFinancingActivities', None) for i in data]
        self.financing_cashflow_items = [i.get('financingCashFlowItems', None) for i in data]
        self.total_cash_dividends_paid = [i.get('totalCashDividendsPaid', None) for i in data]
        self.issuance_retirement_of_stock_net = [i.get('issuanceRetirementofStockNet', None) for i in data]
        self.issuance_retirement_of_debt_net = [i.get('issuanceRetirementofDebtNet', None) for i in data]
        self.foreign_exchange_effects = [i.get('foreignExchangeEffects', None) for i in data]
        self.net_change_in_cash = [i.get('netChangeinCash', None) for i in data]

        self.data_dict = {
            'quoteId': self.quoteid,
            'type': self.type,
            'fiscalYear': self.fiscal_year,
            'fiscalPeriod': self.fiscal_period,
            'endDate': self.end_date,
            'currencyId': self.currency_id,
            'publishDate': self.publish_date,
            'cashfromOperatingActivities': self.cash_from_operating_activities,
            'netIncome': self.net_income,
            'depreciationAndAmortization': self.depreciation_and_amortization,
            'deferredTaxes': self.deferred_taxes,
            'nonCashItems': self.non_cash_items,
            'changesinWorkingCapital': self.changes_in_working_capital,
            'cashfromInvestingActivities': self.cash_from_investing_activities,
            'capitalExpenditures': self.capital_expenditures,
            'otherInvestingCashFlowItemsTotal': self.other_investing_cashflow_items_total,
            'cashfromFinancingActivities': self.cash_from_financing_activities,
            'financingCashFlowItems': self.financing_cashflow_items,
            'totalCashDividendsPaid': self.total_cash_dividends_paid,
            'issuanceRetirementofStockNet': self.issuance_retirement_of_stock_net,
            'issuanceRetirementofDebtNet': self.issuance_retirement_of_debt_net,
            'foreignExchangeEffects': self.foreign_exchange_effects,
            'netChangeinCash': self.net_change_in_cash
        }
      
        self.df = pd.DataFrame(self.data_dict)

    @classmethod
    def from_dict(cls, data):
        return cls(data)

    def __repr__(self):
        return f'<CashFlow quote_id={self.quoteid} fiscal_year={self.fiscal_year} fiscal_period={self.fiscal_period}>'



class BalanceSheet:
    def __init__(self, r):
        data = r['data']

        all_data_dicts = []
    
        self.quoteId = [i.get('quoteId', None) for i in data]
        self.type = [i.get('type', None) for i in data]
        self.fiscalYear = [i.get('fiscalYear', None) for i in data]
        self.fiscalPeriod = [i.get('fiscalPeriod', None) for i in data]
        self.endDate = [i.get('endDate', None) for i in data]
        self.currencyId = [i.get('currencyId', None) for i in data]
        self.publishDate = [i.get('publishDate', None) for i in data]
        self.totalAssets = [i.get('totalAssets', None) for i in data]
        self.totalCurrentAssets = [i.get('totalCurrentAssets', None) for i in data]
        self.cashAndShortTermInvest = [i.get('cashAndShortTermInvest', None) for i in data]
        self.cashEquivalents = [i.get('cashEquivalents', None) for i in data]
        self.shortTermInvestments = [i.get('shortTermInvestments', None) for i in data]
        self.totalReceivablesNet = [i.get('totalReceivablesNet', None) for i in data]
        self.accountsReceTradeNet = [i.get('accountsReceTradeNet', None) for i in data]
        self.totalInventory = [i.get('totalInventory', None) for i in data]
        self.prepaidExpenses = [i.get('prepaidExpenses', None) for i in data]
        self.otherCurrentAssetsTotal = [i.get('otherCurrentAssetsTotal', None) for i in data]
        self.totalNonCurrentAssets = [i.get('totalNonCurrentAssets', None) for i in data]
        self.ppeTotalNet = [i.get('ppeTotalNet', None) for i in data]
        self.ppeTotalGross = [i.get('ppeTotalGross', None) for i in data]
        self.accumulatedDepreciationTotal = [i.get('accumulatedDepreciationTotal', None) for i in data]
        self.otherLongTermAssetsTotal = [i.get('otherLongTermAssetsTotal', None) for i in data]
        self.totalLiabilities = [i.get('totalLiabilities', None) for i in data]
        self.totalCurrentLiabilities = [i.get('totalCurrentLiabilities', None) for i in data]
        self.accountsPayable = [i.get('accountsPayable', None) for i in data]
        self.accruedExpenses = [i.get('accruedExpenses', None) for i in data]
        self.notesPayableShortTermDebt = [i.get('notesPayableShortTermDebt', None) for i in data]
        self.currentPortofLTDebtCapitalLeases = [i.get('currentPortofLTDebtCapitalLeases', None) for i in data]
        self.totalNonCurrentLiabilities = [i.get('totalNonCurrentLiabilities', None) for i in data]
        self.totalLongTermDebt = [i.get('totalLongTermDebt', None) for i in data]
        self.longTermDebt = [i.get('longTermDebt', None) for i in data]
        self.totalDebt = [i.get('totalDebt', None) for i in data]
        self.otherLiabilitiesTotal = [i.get('otherLiabilitiesTotal', None) for i in data]
        self.totalEquity = [i.get('totalEquity', None) for i in data]
        self.totalStockhodersEquity = [i.get('totalStockhodersEquity', None) for i in data]
        self.commonStock = [i.get('commonStock', None) for i in data]
        self.additionalPaidInCapital = [i.get('additionalPaidInCapital', None) for i in data]
        self.retainedEarnings = [i.get('retainedEarnings', None) for i in data]
        self.otherEquityTotal = [i.get('otherEquityTotal', None) for i in data]
        self.totalLiabilitiesShareholdersEquity = [i.get('totalLiabilitiesShareholdersEquity', None) for i in data]
        self.totalCommonSharesOutstanding = [i.get('totalCommonSharesOutstanding', None) for i in data]
        self.data_dict = {
            'quoteId': self.quoteId,
            'type': self.type,
            'fiscalYear': self.fiscalYear,
            'fiscalPeriod': self.fiscalPeriod,
            'endDate': self.endDate,
            'currencyId': self.currencyId,
            'publishDate': self.publishDate,
            'totalAssets': self.totalAssets,
            'totalCurrentAssets': self.totalCurrentAssets,
            'cashAndShortTermInvest': self.cashAndShortTermInvest,
            'cashEquivalents': self.cashEquivalents,
            'shortTermInvestments': self.shortTermInvestments,
            'totalReceivablesNet': self.totalReceivablesNet,
            'accountsReceTradeNet': self.accountsReceTradeNet,
            'totalInventory': self.totalInventory,
            'prepaidExpenses': self.prepaidExpenses,
            'otherCurrentAssetsTotal': self.otherCurrentAssetsTotal,
            'totalNonCurrentAssets': self.totalNonCurrentAssets,
            'ppeTotalNet': self.ppeTotalNet,
            'ppeTotalGross': self.ppeTotalGross,
            'accumulatedDepreciationTotal': self.accumulatedDepreciationTotal,
            'otherLongTermAssetsTotal': self.otherLongTermAssetsTotal,
            'totalLiabilities': self.totalLiabilities,
            'totalCurrentLiabilities': self.totalCurrentLiabilities,
            'accountsPayable': self.accountsPayable,
            'accruedExpenses': self.accruedExpenses,
            'notesPayableShortTermDebt': self.notesPayableShortTermDebt,
            'currentPortofLTDebtCapitalLeases': self.currentPortofLTDebtCapitalLeases,
            'totalNonCurrentLiabilities': self.totalNonCurrentLiabilities,
            'totalLongTermDebt': self.totalLongTermDebt,
            'longTermDebt': self.longTermDebt,
            'totalDebt': self.totalDebt,
            'otherLiabilitiesTotal': self.otherLiabilitiesTotal,
            'totalEquity': self.totalEquity,
            'totalStockhodersEquity': self.totalStockhodersEquity,
            'commonStock': self.commonStock,
            'additionalPaidInCapital': self.additionalPaidInCapital,
            'retainedEarnings': self.retainedEarnings,
            'otherEquityTotal': self.otherEquityTotal,
            'totalLiabilitiesShareholdersEquity': self.totalLiabilitiesShareholdersEquity,
            'totalCommonSharesOutstanding': self.totalCommonSharesOutstanding
        }

          
        self.df = pd.DataFrame(self.data_dict)
    @classmethod
    def from_dict(cls, data):
        return cls(data)

class Forecast:
    def __init__(self, data):
        self.id = data.get('id')
        self.title = data.get('title')
        self.currencyName = data.get('currencyName')
        self.points = data.get('points')

    def __str__(self):
        return f"{self.title}: {self.points}"

    def get_trend(self):
        if len(self.points) < 2:
            return "Unknown"
        
        if self.points[-1].get("valueForecast", 0) > self.points[-2].get("valueForecast", 0):
            return "Increasing"
        else:
            return "Decreasing"


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import aiohttp
class ForecastEvaluator:
    def __init__(self, tickerId, symbol=None):
        self.tickerId = tickerId
        self.mape_values = {}
        self.symbol = symbol

    def calculate_mape(self, actual, forecast):
        actual, forecast = np.array(actual), np.array(forecast)
        mape = np.mean(np.abs((actual - forecast) / actual)) * 100
        return mape

    def plot_metric(self, metric_data, metric_name, report_type):
        dates = [point['xAxis'] for point in metric_data]
        actual_values = [point.get('valueActual', None) for point in metric_data]
        forecast_values = [point.get('valueForecast', None) for point in metric_data]

        fig, ax = plt.subplots()
        ax.plot(dates, actual_values, label='Actual')
        ax.plot(dates, forecast_values, label='Forecast')
        ax.set_xlabel('Date', color='white')
        ax.set_ylabel(metric_name, color='white')
        ax.set_title(f'{metric_name} - Actual vs Forecast {self.symbol}', color='white')
        ax.legend()
        
        ax.spines['bottom'].set_color('white')
        ax.spines['top'].set_color('white')
        ax.spines['left'].set_color('white')
        ax.spines['right'].set_color('white')
        ax.xaxis.label.set_color('white')
        ax.yaxis.label.set_color('white')
        ax.tick_params(axis='x', colors='white')
        ax.tick_params(axis='y', colors='white')

        ax.grid(True, linestyle='--', linewidth=0.5, alpha=0.7)
        plt.xticks(rotation=45)

        # Set background color
        fig.patch.set_facecolor('black')
        ax.set_facecolor('black')
        plt.savefig(f"{metric_name}_{report_type}.png", dpi=300, facecolor='black')
        plt.close()

    async def fetch_data(self, session, url):
        async with session.get(url) as response:
            return await response.json()

    async def get_metric_data(self, statement_type, report_type):
        base_url = f"https://quotes-gw.webullfintech.com/api/information/financial/forecast?tickerId={self.tickerId}&statementType={statement_type}&reportType={report_type}"
        async with aiohttp.ClientSession() as session:
            data = await self.fetch_data(session, base_url)
            return data

    async def evaluate(self):
        report_types = {
            1: 'EPS, EBIT, Net Income, & Revenue',
            2: 'BVPS, Net Asset Value, & Net Debt',
            3: 'Capital Expenditure & CFPS',
            4: 'ROE, ROA',
        }

        for statement_type in report_types.keys():
            for report_type in ['annual', 'quarterly']:
                data = await self.get_metric_data(statement_type, report_type)
                datas = [i['datas'] for i in data]

                for report in reversed(datas):
                    for data in report:
                        id = data['id']
                        title = data['title']
                        points = data['points']
                        actual_forecast_pairs = [(float(point.get('valueActual')), float(point.get('valueForecast'))) for point in points if point.get('valueActual') not in (None, 'N/A') and point.get('valueForecast') not in (None, 'N/A')]
                        if len(actual_forecast_pairs) < 1:
                            continue
                        actual, forecast = zip(*actual_forecast_pairs)
                        actual, forecast = np.array(actual), np.array(forecast)

                        mape = self.calculate_mape(actual, forecast)
                        self.mape_values[title] = mape

                        print(f"ID: {id} TITLE: {title} {report_type.capitalize()} MAPE: {mape}")

                        self.plot_metric(points, title, report_type)





  


from dataclasses import dataclass, asdict
from datetime import datetime

import pandas as pd

@dataclass
class SECFilingsData:
    ticker: str
    filling: str
    description: str
    filling_date: datetime
    report_url: str
    filing_url: str


    def as_dataframe(self) -> pd.DataFrame:
        """Converts the object to a pandas DataFrame."""
        return pd.DataFrame([asdict(self)])

    def insert_into_pg(self, conn):
        cursor = conn.cursor()
        insert_query = """INSERT INTO sec_filings (ticker, filling, description, filling_date, report_url, filing_url) VALUES (%s, %s, %s, %s, %s, %s);"""
        cursor.execute(insert_query, (self.ticker, self.filling, self.description, self.filling_date, self.report_url, self.filing_url))
        conn.commit()



from dataclasses import dataclass, asdict
from datetime import datetime

import pandas as pd


@dataclass
class NewsSentimentData:
    ticker: str
    date: datetime
    title: str
    link: str
    sentiment: str

    def as_dataframe(self) -> pd.DataFrame:
        """Converts the object to a pandas DataFrame."""
        return pd.DataFrame([asdict(self)])

    def insert_into_pg(self, conn):
        cursor = conn.cursor()
        insert_query = """INSERT INTO news_sentiment (ticker, date, title, link, sentiment) VALUES (%s, %s, %s, %s, %s);"""
        cursor.execute(insert_query, (self.ticker, self.date, self.title, self.link, self.sentiment))
        conn.commit()



import os
from dotenv import load_dotenv

load_dotenv()
import random
from fudstop.apis.helpers import map_stock_conditions, STOCK_EXCHANGES, stock_condition_dict, TAPES
from fudstop.apis.helpers import format_large_number, convert_to_ns_datetime
from fudstop.apis.polygonio.async_polygon_sdk import Polygon
from fudstop.list_sets.ticker_lists import most_active_tickers
from polygon.websocket import WebSocketClient, EquityTrade  
from polygon.websocket.models import WebSocketMessage
from typing import List
from discord_webhook import DiscordEmbed, DiscordWebhook
# type: ignore
import orjson
from discord_webhook import AsyncDiscordWebhook, DiscordEmbed
from list_sets.dicts import hex_color_dict
import asyncio
import asyncpg



class StockMarketLive:
    def __init__(self):
        self.pool = None
        
        self.connection_string = os.environ.get('STOCK_MARKET')  # Use one environment variable for connection string
        self.polygon = Polygon(connection_string=self.connection_string)
        self.timespans = ['minute', 'hour', 'day', 'week', 'month']

        self.c = WebSocketClient(subscriptions=["T.*"], custom_json=orjson, api_key=os.environ.get('YOUR_POLYGON_KEY'))

    async def connect(self):
        self.pool = await asyncpg.create_pool(dsn=self.connection_string, min_size=1, max_size=10)
        return self.pool
    async def fetch(self, query):
        async with self.pool.acquire() as conn:
            records = await conn.fetch(query)
            return records
        


    async def fetch_latest_trade(self, ticker):
        async with self.pool.acquire() as connection:
            # The query must have a placeholder for the ticker parameter
            query = "SELECT symbol,size,price,timestamp FROM trades WHERE symbol = $1 ORDER BY timestamp DESC LIMIT 1;"
            latest_trade = await connection.fetchrow(query, ticker)
            return latest_trade
    async def insert_trade(self, trade_data):
        async with self.pool.acquire() as connection:
            async with connection.transaction():
                # The SQL INSERT statement with placeholders for the actual trade data
                insert_query = """
                INSERT INTO trades (
                    event_type, symbol, exchange, external_id, tape, price, size,
                    conditions, timestamp, sequence_number, trf_id, trf_timestamp
                ) VALUES (
                    $1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12
                )
                """
                await connection.execute(
                    insert_query,
                    trade_data['event_type'],
                    trade_data['symbol'],
                    trade_data['exchange'],
                    trade_data['id'],  # Assuming this is the 'external_id'
                    trade_data['tape'],
                    trade_data['price'],
                    trade_data['size'],
                    trade_data['conditions'],  # Ensure this is passed as an array of strings
                    trade_data['timestamp'],
                    trade_data['sequence_number'],
                    trade_data['trf_id'],
                    trade_data['trf_timestamp']
                )

        return self.pool

    async def handle_msg(self, msgs: WebSocketMessage):
        
        
        for message in msgs:

            if hasattr(message, 'conditions') and message.symbol in set(most_active_tickers):
                conditions = [stock_condition_dict.get(i) for i in message.conditions] if message.conditions is not None else []

                exchange = STOCK_EXCHANGES.get(message.exchange)    
                timestamp = convert_to_ns_datetime(message.timestamp)
                trf_timestamp = convert_to_ns_datetime(message.timestamp)

                tape = TAPES.get(message.tape)
                print(timestamp)


                dollar_cost = message.price * message.size


                trade_data = { 

                    'symbol': message.symbol,
                    
                    
                    'event_type': message.event_type,
                    'exchange': exchange,
                    'id': message.id,
                    'tape': tape,
                    'price': message.price,
                    'size': message.size,
                    'conditions': conditions,
                    'timestamp': timestamp,
                    
                    'sequence_number': message.sequence_number,
                    'trf_timestamp': trf_timestamp,
                    'trf_id': message.trf_id,
                    'dollar_cost': dollar_cost,
                    
                    
                }




                await self.insert_trade(trade_data)
                timespan=random.choice(self.timespans)
                rsi_min_task = asyncio.create_task(self.polygon.rsi(ticker=message.symbol, timespan=random.choice(self.timespans)))
     
                rsi = await asyncio.gather(rsi_min_task)
                rsi = rsi[0]
                latest_rsi = None
                if rsi is not None and rsi.rsi_value is not None:
                    latest_rsi = rsi.rsi_value[0]
                    print(latest_rsi)

                if timespan == 'hour':
                    hook = AsyncDiscordWebhook(os.environ.get('hour_osob'))

                if timespan == 'day':
                    hook = AsyncDiscordWebhook(os.environ.get('day_osob'))


                if timespan == 'minute':
                    hook = AsyncDiscordWebhook(os.environ.get('minute_osob'))

                if timespan == 'week':
                    hook = AsyncDiscordWebhook(os.environ.get('week_osob'))

                if timespan == 'month':
                    hook = AsyncDiscordWebhook(os.environ.get('month_osob'))

                
                if timespan is not None and hook is not None and latest_rsi is not None:
                    color = hex_color_dict['green'] if latest_rsi <=30 else hex_color_dict['red'] if latest_rsi >= 70 else hex_color_dict['grey']
                    status = 'oversold' if color == hex_color_dict['green'] else 'overbought' if color == hex_color_dict['red'] else None
                    if status is not None:
                        try:
                            embed = DiscordEmbed(title=f"RSI Results - ALL {timespan}", description=f"```py\n{message.symbol} is {status} on the {timespan} with an RSI value of {latest_rsi}.```", color=color)
                            embed.set_timestamp()
                            embed.set_footer(f'overbought / oversold RSI feed - {timespan}')
                            hook.add_embed(embed)
                            await hook.execute()
                        except TypeError:
                            continue

market = StockMarketLive()

async def main():
    await market.connect()
    await asyncio.gather(market.c.connect(market.handle_msg))
    

if __name__ == '__main__':
    asyncio.run(main())



import os

import pandas as pd

class DocumentQuery:
    def __init__(self, data):


        self.count = data.get('count')
        self.description = data.get('description')
        self.total_pages = data.get('total_pages')
        results = data.get('results')

        self.title = [i.get('title', 'N/A') for i in results]
        self.type = [i.get('type', 'N/A') for i in results]
        self.abstract = [i.get('abstract', 'N/A') for i in results]
        self.document_number = [i.get('document_number', 'N/A') for i in results]
        self.html_url= [i.get('html_url', 'N/A') for i in results]
        self.pdf_url= [i.get('pdf_url', 'N/A') for i in results]
        self.public_inspection_pdf_url= [i.get('public_inspection_pdf_url', 'N/A') for i in results]
        self.publication_date= [i.get('publication_date', 'N/A') for i in results]
        agencies= [i.get('agencies', 'N/A') for i in results]
        # Assuming agencies[0][0] is a dictionary
        attributes = ['raw_name', 'name', 'id', 'url', 'json_url', 'parent_id', 'slug']
        extracted_attributes = [{attr: agency[0].get(attr, 'N/A') for attr in attributes} for agency in agencies]
        # Extract only the 'name' attribute from each agency dictionary
        self.agency_names = [agency['name'] for agency in extracted_attributes]

        self.excerpts= [i.get('excerpts', 'N/A') for i in results]
        print(extracted_attributes)
 


        self.data_dict = { 

            'result_count': self.count,
            'search_description': self.description,
            'total_pages': self.total_pages,
            'title': self.title,
            'type': self.type,
            'abstract': self.abstract,
            'document_number': self.document_number,
            'html_url': self.html_url,
            'pdf_url': self.pdf_url,
            'public_inspection_url': self.public_inspection_pdf_url,
            'publication_date': self.publication_date,
            'excerpts': self.excerpts,
            'agency': self.agency_names

        }
      

        self.as_dataframe = pd.DataFrame(self.data_dict)

        # Create the directory if it doesn't exist
        directory = 'data/fed_register/'
        if not os.path.exists(directory):
            os.makedirs(directory)


        # Create the initial DataFrame
        self.as_dataframe = pd.DataFrame(self.data_dict)

        # Create directory if it doesn't exist
        directory = 'data/fed_register/'
        if not os.path.exists(directory):
            os.makedirs(directory)

        # Save DataFrame to CSV
        self.as_dataframe.to_csv('data/fed_register/query_results.csv')


from dataclasses import dataclass, asdict
from datetime import datetime

import pandas as pd

@dataclass
class ShortVolumeData:
    date: datetime
    ticker: str
    short_vol: int
    short_exempt_vol: int
    total_vol: int
    percent_shorted: float

    def as_dataframe(self) -> pd.DataFrame:
        """Converts the object to a pandas DataFrame."""
        return pd.DataFrame([asdict(self)])

    def insert_into_pg(self, conn):
        cursor = conn.cursor()
        insert_query = """INSERT INTO short_volume (date, ticker, short_vol, short_exempt_vol, total_vol, percent_shorted) VALUES (%s, %s, %s, %s, %s, %s);"""
        cursor.execute(insert_query, (self.date, self.ticker, self.short_vol, self.short_exempt_vol, self.total_vol, self.percent_shorted))
        conn.commit()



from apis.polygonio.async_polygon_sdk import Polygon
import os

import asyncio

poly = Polygon()


# Creating a database configuration dictionary using os.environ.get for environment variables
db_config = {
    "DB_HOST": os.environ.get('DB_HOST', 'localhost'),
    "DB_PASSWORD": os.environ.get('DB_PASSWORD'),
    "DB_NAME": os.environ.get('COMPANY_INFO_STRING'),
    "DB_PORT": os.environ.get('DB_PORT', 5432),
    "DB_USER": os.environ.get('DB_USER')
}



async def ticker_snapshot():

    ticker_data = await poly.get_all_tickers(save_all_tickers=True)


    print(f"Number of tickers found: {len(ticker_data)}")

    #easily print attributes / work with attributes using dot notation:

    for attribute in ticker_data:


        #printing the parents
        print(f"Last Trade:",attribute.last_trade)
        print(f"Last Quote:",attribute.stock_last_quote)
        print(f"Day:",attribute.stock_day)
        print(f"Prev Day:",attribute.prev_day)
        print(f"Minute:", attribute.stock_minute_bar)


        #printing the children within a parent

        """
        >>> Last trade example:
        """

        last_trade = attribute.last_trade.conditions
        print(last_trade)


asyncio.run(ticker_snapshot())


import os



class DiscordSDK:
    def __init__(self):
        """
        A toolkit for working with the discord API.

        Your token is retrieved from the developer tools window. 

        >>> Network tab:

        Type a message in a channel while the dev. tools window is open with the 
        "Network" tab selected and filtered by "FETCH/XHR". 


        >>> URL:

        After typing the message - you'll see a URL pop up. Click on it - and scroll 
        down until you see the "Request Headers" section.
        
        Look for "authorization" and copy the token beside it, and paste it into an .env file as YOUR_DISCORD_HTTP_TOKEN=.
        """


        self.token = os.environ.get('YOUR_DISCORD_HTTP_TOKEN')


from setuptools import setup, find_packages

# Read requirements.txt and remove any comments
with open('requirements.txt') as f:
    requirements = f.read().splitlines()
    requirements = [r.strip() for r in requirements if not r.startswith('#')]

setup(
    name='fudstop',
    version='0.1.8',
    packages=find_packages(),
    install_requires=requirements,
)


import os
import sys
import pandas as pd
from pathlib import Path
import aiohttp
import asyncio
from asyncpg.exceptions import UniqueViolationError
from asyncpg import create_pool
from dotenv import load_dotenv
load_dotenv()
# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[1])
if project_dir not in sys.path:
    sys.path.append(project_dir)
from urllib.parse import urlencode
from datetime import datetime, timedelta
from webull.webull_trading import WebullTrading
from .models.option_models.universal_snapshot import UniversalOptionSnapshot
from .models.option_models.option_snapshot import OptionSnapshotData
from .polygon_helpers import flatten_nested_dict, flatten_dict

trading = WebullTrading()

class PolygonOptions:
    def __init__(self, connection_string=None):
        self.connection_string = connection_string
        self.host = os.environ.get('DB_HOST')
        self.port = os.environ.get('DB_PORT')
        self.user = os.environ.get('DB_USER')
        self.password = os.environ.get('DB_PASSWORD')
        self.database = os.environ.get('POLYGON')

        self.api_key = os.environ.get('YOUR_POLYGON_KEY')
        self.today = datetime.now().strftime('%Y-%m-%d')
        self.yesterday = (datetime.now() - timedelta(days=1)).strftime('%Y-%m-%d')
        self.tomorrow = (datetime.now() + timedelta(days=1)).strftime('%Y-%m-%d')
        self.thirty_days_ago = (datetime.now() - timedelta(days=30)).strftime('%Y-%m-%d')
        self.thirty_days_from_now = (datetime.now() + timedelta(days=30)).strftime('%Y-%m-%d')
        self.fifteen_days_ago = (datetime.now() - timedelta(days=15)).strftime('%Y-%m-%d')
        self.fifteen_days_from_now = (datetime.now() + timedelta(days=15)).strftime('%Y-%m-%d')
        self.eight_days_from_now = (datetime.now() + timedelta(days=8)).strftime('%Y-%m-%d')
        self.eight_days_ago = (datetime.now() - timedelta(days=8)).strftime('%Y-%m-%d')
        self.one_year_from_now = (datetime.now() + timedelta(days=365)).strftime('%Y-%m-%d')
        self.one_year_ago = (datetime.now() - timedelta(days=365)).strftime('%Y-%m-%d')
    async def connect(self, connection_string=None):
            if connection_string:
                self.pool = await create_pool(
                    dsn=connection_string, min_size=1, max_size=10
                )
            else:
                self.pool = await create_pool(
                    host=self.host,
                    port=self.port,
                    user=self.user,
                    password=self.password,
                    database=self.database,
                    min_size=1,
                    max_size=10
                )
            return self.pool

    async def save_structured_messages(self, data_list: list[dict], table_name: str):
        if not data_list:
            return  # No data to insert

        # Assuming all dicts in data_list have the same keys
        fields = ', '.join(data_list[0].keys())
        values_placeholder = ', '.join([f"${i+1}" for i in range(len(data_list[0]))])
        values = ', '.join([f"({values_placeholder})" for _ in data_list])
        
        query = f'INSERT INTO {table_name} ({fields}) VALUES {values}'
        print(self.connection_string)
        async with self.pool.acquire() as conn:
            try:
                flattened_values = [value for item in data_list for value in item.values()]
                await conn.execute(query, *flattened_values)
            except UniqueViolationError:
                print('Duplicate - Skipping')


    async def paginate_concurrent(self, url, as_dataframe=False, concurrency=25):
        all_results = []

        
        async with aiohttp.ClientSession() as session:
            pages_to_fetch = [url]
            
            while pages_to_fetch:
                tasks = []
                
                for _ in range(min(concurrency, len(pages_to_fetch))):
                    next_url = pages_to_fetch.pop(0)
                    tasks.append(self.fetch_page(next_url))
                    
                results = await asyncio.gather(*tasks)
                if results is not None:
                    for data in results:
                        if data is not None:
                            if "results" in data:
                                all_results.extend(data["results"])
                                
                            next_url = data.get("next_url")
                            if next_url:
                                next_url += f'&{urlencode({"apiKey": f"{self.api_key}"})}'
                                pages_to_fetch.append(next_url)
                        else:
                            break
        if as_dataframe:
            import pandas as pd
            return pd.DataFrame(all_results)
        else:
            return all_results
        



    #

    async def fetch_page(self, url):
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                response.raise_for_status()
                return await response.json()
    

    async def fetch_endpoint(self, endpoint, params=None):
        """
        Uses endpoint / parameter combinations to fetch the polygon API

        dynamically.
        
        """
        # Build the query parameters
        filtered_params = {k: v for k, v in params.items() if v is not None} if params else {}
        
        if filtered_params:
            query_string = urlencode(filtered_params)
            if query_string:
                endpoint = f"{endpoint}?{query_string}&{self.api_key}"
                
        async with aiohttp.ClientSession() as session:
            async with session.get(url=endpoint, params=filtered_params) as response:
                response_data = await response.json()
                if 'next_url' in response_data:
                    # If "next_url" exists - the function auto paginates
                    return await self.paginate_concurrent(endpoint,as_dataframe=True,concurrency=40)
                return response_data  # or response.text() based on your needs
                
    async def get_price(self, ticker:str):
        """
        Fetches price from Webull API to use for option queries

        Arguments:

        >>> ticker: required - the ticker to survey
        """

        if ticker.startswith('I:'):
            ticker = ticker.replace('I:', '')
        datas = await trading.stock_quote(ticker)

        price = datas.web_stock_close

        return price

    async def get_option_contracts(self, ticker:str, 
                                   strike_price_greater_than:str=None, 
                                   strike_price_less_than:str=None, 
                                   expiry_date_greater_than:str=None, 
                                   expiry_date_less_than:str=None, limit:str='250'):
        """
        Returns options for a specified ticker.

        Arguments:

        >>> ticker: the ticker to query (required)

        >>> strike_price_greater_than: the minimum strike to be returned (optional)      


        >>> strike_price_less_than: the maximum strike to be returned (optional) 


        >>> expiry_greater_than: the minimum expiry date to be returned (optional)


        >>> expiry_less_than: the maximum expiry date to be returned (optional)

        >>> limit: the amount of contracts to be returned (max 250 per request)

        
        """


        endpoint = f"https://api.polygon.io/v3/snapshot/options/{ticker}"

        
        params = {
            "limit": limit,
            "apiKey": self.api_key
        }

        df = await self.fetch_endpoint(endpoint, params)
      
        # Columns to normalize
        columns_to_normalize = ['day', 'underlying_asset', 'details', 'greeks', 'last_quote', 'last_trade']

        # Apply the helper function to each row and each specified column
        for column in columns_to_normalize:
            df = df.apply(lambda row: flatten_nested_dict(row, column), axis=1)

        return df

    async def get_strike_thresholds(self, ticker:str, price):
        indices_list = ["SPX", "SPXW", "NDX", "VIX", "VVIX"]
        if price is not None and ticker in indices_list:
            lower_strike = round(float(price) * 0.97)
            upper_strike = round(float(price) * 1.03)
            return lower_strike, upper_strike
        else:
            lower_strike = round(float(price) * 0.85)
            upper_strike = round(float(price) * 1.15)
            return lower_strike, upper_strike

    async def get_near_the_money_options(self, ticker: str, exp_greater_than:str=None, exp_less_than:str=None):
        """
        Gets options near the money for a ticker
        """

        if exp_greater_than is None:
            exp_greater_than = self.today

        if exp_less_than is None:
            exp_less_than = self.eight_days_from_now



        price = await self.get_price(ticker)
  
        if price is not None:
            upper_strike, lower_strike = await self.get_strike_thresholds(ticker, price)
      
            async with aiohttp.ClientSession() as session:
                url = f"https://api.polygon.io/v3/snapshot/options/{ticker}?strike_price.lte={lower_strike}&strike_price.gte={upper_strike}&expiration_date.gte={exp_greater_than}&expiration_date.lte={exp_less_than}&limit=250&apiKey={self.api_key}"
         
                async with session.get(url) as resp:
                    r = await resp.json()
                    results = r['results'] if 'results' in r else None
                    if results is None:
                        return
                    else:
                        results = UniversalOptionSnapshot(results)
                       
                        tickers = results.ticker
                        if ticker is not None:
                            atm_tickers = ','.join(tickers)
                            return atm_tickers
                        else:
                            return None
                        
    async def get_near_the_money_data(self, ticker: str, exp_greater_than:str='2023-11-17', exp_less_than:str='2023-12-15'):
        """
        Gets options near the money for a ticker
        """





        price = await self.get_price(ticker)
  
        if price is not None:
            upper_strike, lower_strike = await self.get_strike_thresholds(ticker, price)
      
            async with aiohttp.ClientSession() as session:
                url = f"https://api.polygon.io/v3/snapshot/options/{ticker}?strike_price.lte={lower_strike}&strike_price.gte={upper_strike}&expiration_date.gte={exp_greater_than}&expiration_date.lte={exp_less_than}&limit=250&apiKey={self.api_key}"
         
                data = await self.paginate_concurrent(url)

                return UniversalOptionSnapshot(data).df
            
    async def get_skew(self, ticker:str):
        """Gets the IV skew across all expiration dates along with the strikes above and below the lowest IV.
        
        Arguments:
        >>> ticker: required - the ticker to survey
        """

        options = await self.get_option_contracts(ticker, expiry_date_less_than='2023-11-03')

        # Sort by 'expiration_date' and 'implied_volatility'
        options_sorted = options.sort_values(by=['expiration_date', 'implied_volatility'])

        results = []

        # For each expiration date, get the lowest IV strike and the ones above and below it
        for expiration, group in options_sorted.groupby('expiration_date'):
            min_iv_idx = group['implied_volatility'].idxmin()
            
            # Get rows for lowest IV and the ones immediately above and below it
            rows = group.loc[[min_iv_idx - 1, min_iv_idx, min_iv_idx + 1], :]
            
            results.append(rows)

        # Concatenate all the results
        result_df = pd.concat(results)

     
        return result_df


    async def get_universal_snapshot(self, ticker, retries=3): #✅
        """Fetches the Polygon.io universal snapshot API endpoint"""
        timeout = aiohttp.ClientTimeout(total=10)  # 10 seconds timeout for the request
        
        for retry in range(retries):
        # async with sema:
            url = f"https://api.polygon.io/v3/snapshot?ticker.any_of={ticker}&apiKey={self.api_key}&limit=250"

            
            async with aiohttp.ClientSession(timeout=timeout) as session:
                try:
                    async with session.get(url) as resp:
                        data = await resp.json()
                        results = data.get('results', None)
        
                        if results is not None:
                            flattened_results = [flatten_dict(result) for result in results]
                            return flattened_results
                            
                except aiohttp.ClientConnectorError:
                    print("ClientConnectorError occurred. Retrying...")
                    continue
                
                except aiohttp.ContentTypeError as e:
                    print(f"ContentTypeError occurred: {e}")  # Consider logging this
                    continue
                
                except Exception as e:
                    print(f"An unexpected error occurred: {e}")  # Consider logging this
                    continue


    async def get_option_chain_all(self, underlying_asset, strike_price=None, expiration_date=None, contract_type=None, order=None, limit=250, sort=None):
        """
        Get all options contracts for an underlying ticker across all pages.

        :param underlying_asset: The underlying ticker symbol of the option contract.
        :param strike_price: Query by strike price of a contract.
        :param expiration_date: Query by contract expiration with date format YYYY-MM-DD.
        :param contract_type: Query by the type of contract.
        :param order: Order results based on the sort field.
        :param limit: Limit the number of results returned, default is 10 and max is 250.
        :param sort: Sort field used for ordering.
        :return: A list containing all option chain data across all pages.
        """
        endpoint = f"https://api.polygon.io/v3/snapshot/options/{underlying_asset}?limit=250&apiKey={self.api_key}"

        async with aiohttp.ClientSession() as session:
            response_data = await self.paginate_concurrent(endpoint)
            print(endpoint) #passing endpoint to paginate_concurrent
            option_data = OptionSnapshotData(response_data)

            return option_data


import os
from dotenv import load_dotenv
import pandas as pd
load_dotenv()

import asyncio
from datetime import datetime
from fudstop.apis.webull.webull_options import WebullOptions

options = WebullOptions(os.environ.get('WEBULL_OPTIONS'))



import disnake
from disnake.ext import commands

import disnake
import asyncpg

class OptionModal(disnake.ui.Modal):
    def __init__(self):
        components = [
            disnake.ui.TextInput(
                label="Symbol",
                custom_id="input_1",
                style=disnake.TextInputStyle.short,
                max_length=7,
                placeholder='e.g. SPY'
            ),
            disnake.ui.TextInput(
                label="Expiration",
                placeholder='e.g. 2023-12-15',
                max_length=10,
                custom_id="input_2",
                style=disnake.TextInputStyle.short,
            ),
            disnake.ui.TextInput(
                label="Call or Put?",
                custom_id="input_3",
                style=disnake.TextInputStyle.short,
            ),
            disnake.ui.TextInput(
                label="What Strike?",
                custom_id="input_4",
                placeholder='e.g. 125.5',
                style=disnake.TextInputStyle.short,
            ),
        ]
        super().__init__(title="Option Data Query", custom_id="option_query_modal", components=components)

    async def callback(self, inter: disnake.ModalInteraction):
        await options.connect()
        symbol = inter.text_values['input_1']
        expiration = inter.text_values['input_2']
        call_put = inter.text_values['input_3'].lower()
        strike = inter.text_values['input_4']

        # Ensure that the strike is a float and expiration is a valid date string
    # Convert the strike to a float and the expiration to a date object
        try:
            strike = float(strike)
            expiration = datetime.strptime(expiration, '%Y-%m-%d').date()
        except ValueError:
            await inter.response.send_message("Invalid strike price or expiration date format.", ephemeral=True)
            return

        # Placeholder for database connection
        # conn = await asyncpg.connect('postgresql://user:password@host/database')
        
        query = f"""
        SELECT * FROM options_data
        WHERE symbol = $1
        AND expiry_date = $2
        AND call_put = $3
        AND strike_price = $4
        """
        
        
        records = await options.fetch(query, symbol, expiration, call_put, strike)
    

   
        await inter.response.send_message(f"Query results: {records}")


        print(query)
        print(symbol, expiration, call_put, strike)

        # After performing the query, you would process the records and send back a response
        # For example:
        # formatted_records = process_records(records)
        # await inter.response.send_message(formatted_records, ephemeral=True)


class SQLQueryModal(disnake.ui.Modal):
    def __init__(self):
        components = [
            disnake.ui.TextInput(
                label="Columns",
                custom_id="input_1",
                style=disnake.TextInputStyle.short,
                placeholder='e.g. column1, column2, ...'
            ),
            disnake.ui.TextInput(
                label="WHERE",
                custom_id="input_2",
                style=disnake.TextInputStyle.short,
                placeholder='e.g. column_name = value'
            ),
            disnake.ui.TextInput(
                label="ORDER BY",
                custom_id="input_3",
                style=disnake.TextInputStyle.short,
                placeholder='e.g. column_name ASC/DESC'
            ),
            disnake.ui.TextInput(
                label="LIMIT",
                custom_id="input_4",
                style=disnake.TextInputStyle.short,
                placeholder='e.g. 10'
            ),
        ]
        super().__init__(title="SQL Query Constructor", custom_id="sql_query_modal", components=components)

    async def callback(self, inter: disnake.ModalInteraction):
        await options.connect()
        await inter.response.defer()
        columns = inter.text_values['input_1']
        where = inter.text_values['input_2']
        order_by = inter.text_values['input_3']
        limit = inter.text_values['input_4']

        # Construct the query
        query = f"SELECT {columns} FROM options_data"
        if where:
            query += f" WHERE {where}"
        if order_by:
            query += f" ORDER BY {order_by}"
        if limit:
            query += f" LIMIT {limit}"

        # Send the constructed query back to the user
        await inter.edit_original_message(f"Constructed SQL Query: ```sql\n{query}\n```")
        await asyncio.sleep(2)
        await inter.edit_original_message(f"> # Querying Database..")
        await asyncio.sleep(2)
        await inter.edit_original_message(f"> # Building CSV file...")

        await asyncio.sleep(4)
        results = await options.fetch(query)    

        df = pd.DataFrame(results)

        df.to_csv('data_query.csv')
        await inter.edit_original_message(file=disnake.File(f"data_query.csv"))
# In your bot command, you would invoke this modal like so:


import asyncio
import os

import aiohttp

from dotenv import load_dotenv

key = os.environ.get('YOUR_UNSPLASH_ACCESS_KEY')

from .models.unsplash_models import ImageURLS

class Unsplash:
    def __init__(self):
        self.key = key



    async def get_image(self, query:str):
        """
        Fetches the UNSPLASH API and returns an image based on user query.


        Arguments:

        >>> query: the image to query
        
        """

        url = f"https://api.unsplash.com/search/photos?page=1&query={query}&client_id={self.key}"

        async with aiohttp.ClientSession() as session:
            async with session.get(url) as resp:
                data = await resp.json()
                results = data['results']
                urls = [i.get('urls') for i in results]

                urls = urls[0]

                return ImageURLS(urls)










option_condition_dict = {
    201: 'Canceled',
    202: 'Late and Out Of Sequence',
    203: 'Last and Canceled',
    204: 'Late',
    205: 'Opening Trade and Canceled',
    206: 'Opening Trade, Late, and Out Of Sequence',
    207: 'Only Trade and Canceled',
    208: 'Opening Trade and Late',
    209: 'Automatic Execution',
    210: 'Reopening Trade',
    219: 'Intermarket Sweep Order',
    227: 'Single Leg Auction Non ISO',
    228: 'Single Leg Auction ISO',
    229: 'Single Leg Cross Non ISO',
    230: 'Single Leg Cross ISO',
    231: 'Single Leg Floor Trade',
    232: 'Multi Leg auto-electronic trade',
    233: 'Multi Leg Auction',
    234: 'Multi Leg Cross',
    235: 'Multi Leg floor trade',
    236: 'Multi Leg auto-electronic trade against single leg(s)',
    237: 'Stock Options Auction',
    238: 'Multi Leg Auction against single leg(s)',
    239: 'Multi Leg floor trade against single leg(s)',
    240: 'Stock Options auto-electronic trade',
    241: 'Stock Options Cross',
    242: 'Stock Options floor trade',
    243: 'Stock Options auto-electronic trade against single leg(s)',
    244: 'Stock Options Auction against single leg(s)',
    245: 'Stock Options floor trade against single leg(s)',
    246: 'Multi Leg Floor Trade of Proprietary Products',
    247: 'Multilateral Compression Trade of Proprietary Products',
    248: 'Extended Hours Trade',
}

quote_conditions = {
    -1: "Invalid",
    0: "Regular",
    1: "RegularTwoSidedOpen",
    2: "RegularOneSidedOpen",
    3: "SlowAsk",
    4: "SlowBid",
    5: "SlowBidAsk",
    6: "SlowDueLRPBid",
    7: "SlowDueLRPAsk",
    8: "SlowDueNYSELRP",
    9: "SlowDueSetSlowListBidAsk",
    10: "ManualAskAutomatedBid",
    11: "ManualBidAutomatedAsk",
    12: "ManualBidAndAsk",
    13: "Opening",
    14: "Closing",
    15: "Closed",
    16: "Resume",
    17: "FastTrading",
    18: "TradingRangeIndication",
    19: "MarketMakerQuotesClosed",
    20: "NonFirm",
    21: "NewsDissemination",
    22: "OrderInflux",
    23: "OrderImbalance",
    24: "DueToRelatedSecurityNewsDissemination",
    25: "DueToRelatedSecurityNewsPending",
    26: "AdditionalInformation",
    27: "NewsPending",
    28: "AdditionalInformationDueToRelatedSecurity",
    29: "DueToRelatedSecurity",
    30: "InViewOfCommon",
    31: "EquipmentChangeover",
    32: "NoOpenNoResponse",
    33: "SubPennyTrading",
    34: "AutomatedBidNoOfferNoBid",
    35: "LULDPriceBand",
    36: "MarketWideCircuitBreakerLevel1",
    37: "MarketWideCircuitBreakerLevel2",
    38: "MarketWideCircuitBreakerLevel3",
    39: "RepublishedLULDPriceBand",
    40: "OnDemandAuction",
    41: "CashOnlySettlement",
    42: "NextDaySettlement",
    43: "LULDTradingPause",
    71: "SLowDuelRPBidAsk",
    80: "Cancel",
    81: "Corrected_Price",
    82: "SIPGenerated",
    83: "Unknown",
    84: "Crossed_Market",
    85: "Locked_Market",
    86: "Depth_On_Offer_Side",
    87: "Depth_On_Bid_Side",
    88: "Depth_On_Bid_And_Offer",
    89: "Pre_Opening_Indication",
    90: "Syndicate_Bid",
    91: "Pre_Syndicate_Bid",
    92: "Penalty_Bid"
}
option_condition_desc_dict = {
    
    "Canceled": "Transaction previously reported (other than as the last or opening report for the particular option contract) is now to be canceled.",
    "Late and Out Of Sequence": "Transaction is being reported late and is out of sequence; i.e., later transactions have been reported for the particular option contract.",
    "Last and Canceled": "Transaction is the last reported for the particular option contract and is now canceled.",
    "Late": "Transaction is being reported late, but is in the correct sequence; i.e., no later transactions have been reported for the particular option contract.",
    "Opening Trade and Canceled": "Transaction was the first one (opening) reported this day for this particular option contract. Although later transactions have been reported, this transaction is now to be canceled.",
    "Opening Trade, Late, and Out Of Sequence": "Transaction is a late report of the opening trade and is out of sequence; i.e., other transactions have been reported for the particular option contract.",
    "Only Trade and Canceled": "Transaction was the only one reported this day for the particular option contract and is now to be canceled.",
    "Opening Trade and Late": "Transaction is a late report of the opening trade, but is in the correct sequence; i.e., no other transactions have been reported for the particular option contract.",
    "Automatic Execution": "Transaction was executed electronically. Prefix appears solely for information; process as a regular transaction.",
    "Reopening Trade": "Transaction is a reopening of an option contract in which trading has been previously halted. Prefix appears solely for information; process as a regular transaction.",
    "Intermarket Sweep Order": "Transaction was the execution of an order identified as an Intermarket Sweep Order. Process like normal transaction.",
    "Single Leg Auction Non ISO": "Transaction was the execution of an electronic order which was 'stopped' at a price and traded in a two-sided auction mechanism that goes through an exposure period. Such auctions mechanisms include and are not limited to Price Improvement, Facilitation, or Solicitation Mechanism.",
    "Single Leg Auction ISO": "Transaction was the execution of an Intermarket Sweep electronic order which was 'stopped' at a price and traded in a two-sided auction mechanism that goes through an exposure period. Such auctions mechanisms include and are not limited to Price Improvement, Facilitation, or Solicitation Mechanism marked as ISO.",
    "Single Leg Cross Non ISO": "Transaction was the execution of an electronic order which was 'stopped' at a price and traded in a two-sided crossing mechanism that does not go through an exposure period. Such crossing mechanisms include and are not limited to Customer to Customer Cross and QCC with a single option leg.",
    "Single Leg Cross ISO": "Transaction was the execution of an Intermarket Sweep electronic order which was 'stopped' at a price and traded in a two-sided crossing mechanism that does not go through an exposure period. Such crossing mechanisms include and are not limited to Customer to Customer Cross.",
    "Single Leg Floor Trade": "Transaction represents a non-electronic trade executed on a trading floor. Execution of Paired and Non-Paired Auctions and Cross orders on an exchange floor are also included in this category.",
    "Multi Leg auto-electronic trade": "Transaction represents an electronic execution of a multi-leg order traded in a complex order book.",
    "Multi Leg Auction": "Transaction was the execution of an electronic multi-leg order which was stopped at a price and traded in a two-sided auction mechanism that goes through an exposure period in a complex order book. Such auctions mechanisms include and are not limited to Price Improvement, Facilitation, or Solicitation Mechanism.",
    "Multi Leg Cross": "Transaction was the execution of an electronic multi-leg order which was 'stopped' at a price and traded in a two-sided crossing mechanism that does not go through an exposure period. Such crossing mechanisms include and are not limited to Customer to Customer Cross and QCC with two or more options legs.",
    "Multi Leg floor trade": "Transaction represents a non-electronic multi-leg order trade executed against other multi-leg order(s) on a trading floor. Execution of Paired and Non-Paired Auctions and Cross orders on an exchange floor are also included in this category.",
    "Multi Leg auto-electronic trade against single leg(s)": "Transaction represents an electronic execution of a multi-leg order traded against single leg orders/quotes.",
    "Stock Options Auction": "Transaction was the execution of an electronic multi-leg stock/options order which was 'stopped' at a price and traded in a two-sided auction mechanism that goes through an exposure period in a complex order book. Such auctions mechanisms include and are not limited to Price Improvement, Facilitation, or Solicitation Mechanism.",
    "Multi Leg Auction against single leg(s)": "Transaction was the execution of an electronic multi-leg order which was 'stopped' at a price and traded in a two-sided auction mechanism that goes through an exposure period and trades against single leg orders/quotes. Such auctions mechanisms include and are not limited to Price Improvement, Facilitation, or Solicitation Mechanism.",
    "Multi Leg floor trade against single leg(s)": "Transaction represents a non-electronic multi-leg order trade executed on a trading floor against single leg orders/quotes. Such auctions mechanisms include and are not limited to Price Improvement, Facilitation, or Solicitation Mechanism.",
    "Stock Options auto-electronic trade": "Transaction represents the an electronic execution of a multi-leg stock/options order traded in a complex order book.",
    "Stock Options Cross": "Transaction was the execution of an electronic multi-leg stock/options order which was 'stopped' at a price and traded in a two-sided crossing mechanism that does not go through an exposure period. Such crossing mechanisms include and are not limited to Customer to Customer Cross.",
    "Stock Options floor trade": "Transaction represents a non-electronic multi-leg order stock/options trade executed on a trading floor in a complex order book. Execution of Paired and Non-Paired Auctions and Cross orders on an exchange floor are also included in this category.",
    "Stock Options auto-electronic trade against single leg(s)": "Transaction represents an electronic execution of a multi-leg stock/options order traded against single leg orders/quotes.",
    "Stock Options Auction against single leg(s)": "Transaction was the execution of an electronic multi-leg stock/options order which was 'stopped' at a price and traded in a two-sided auction mechanism that goes through an exposure perod and trades against single leg orders/quotes. Such auctions mechanisms include and are not limited to Price Improvement, Facilitation, or Solicitation Mechanism.",
    "Stock Options floor trade against single leg(s)": "Transaction represents a non-electronic multi-leg stock/options order trade executed on a trading floor against single leg orders/quotes. Execution of Paired and Non-Paired Auctions on an exchange floor are also included in this category.",
    "Multi Leg Floor Trade of Proprietary Products": "Transaction represents execution of a proprietary product non-electronic multi-leg order with at least 3 legs. The trade price may be outside the current NBBO.",
    "Multilateral Compression Trade of Proprietary Products": "Transaction represents an execution in a proprietary product done as part of a multilateral compression. Trades are executed outside of regular trading hours at prices derived from end of day markets. Trades do not update Open High, Low, and Closing Prices.",
    "Extended Hours Trade": "Transaction represents a trade that was executed outside of regular market hours. Trades do not update Open, High, Low, and Closing Prices."
}

stock_condition_dict = {
    0: 'Regular Sale',
    1: 'Acquisition',
    2: 'Average Price Trade',
    3: 'Automatic Execution',
    4: 'Bunched Trade',
    5: 'Bunched Sold Trade',
    7: 'Cash Sale',
    8: 'Closing Prints',
    9: 'Cross Trade',
    10: 'Derivatively Priced',
    11: 'Distribution',
    12: 'Form T',
    13: 'Extended Trading Hours (Sold Out of Sequence)',
    14: 'Intermarket Sweep',
    15: 'Market Center Official Close',
    16: 'Market Center Official Open',
    17: 'Market Center Opening Trade',
    18: 'Market Center Reopening Trade',
    19: 'Market Center Closing Trade',
    20: 'Next Day',
    21: 'Price Variation Trade',
    22: 'Prior Reference Price',
    23: 'Rule 155 Trade (AMEX)',
    24: 'Rule 127 NYSE',
    25: 'Opening Prints',
    27: 'Stopped Stock (Regular Trade)',
    28: 'Re-Opening Prints',
    29: 'Seller',
    30: 'Sold Last',
    33: 'Sold (out of Sequence)',
    34: 'Split Trade',
    35: 'Stock Option Trade',
    36: 'Yellow Flag Regular Trade',
    37: 'Odd Lot Trade',
    38: 'Corrected Consolidated Close (per listing market)',
    40: 'Held',
    41: 'Trade Thru Exempt',
    46: 'Contingent Trade',
    52: 'Contingent Trade',
    53: 'Qualified Contingent Trade',
    59: 'Placeholder for 611 exempt',
    60: 'SSR in Effect',}

stuff={'Corrected Consolidated Close (per listing market)': 'Corrected Consolidated Close (per listing market)',
    'Cash Only Settlement': 'Cash Only Settlement',
    'Closing Prints': 'Closing Prints',
    'Slow Due Set Slow List Bid Ask': 'Slow Due Set Slow List Bid Ask',
    'Manual Bid and Ask': 'Manual Bid and Ask',
    'Multi Leg auto-electronic trade': 'Multi Leg auto-electronic trade',
    'Single Leg Auction Non ISO': 'Single Leg Auction Non ISO',
    'Multi Leg Auction': 'Multi Leg Auction',
    'Automatic Execution': 'Automatic Execution',
}

stock_condition_desc_dict = {
    'Regular Sale': 'Regular Sale - A trade made without stated conditions is deemed regular way for settlement on the third business day following the transaction date.',
    'Financially Deficient': 'Financially Deficient - Company doesnt meet minimum listing requirements',
    'Acquisition': 'Acquisition - A transaction made on the Exchange as a result of an Exchange acquisition.',
    'Average Price Trade': 'Average Price Trade - A trade where the price reported is based upon an average of the prices for transactions in a security during all or any portion of the trading day.',
    'Automatic Execution': 'Automatic Execution - A sale condition code that identifies a NYSE trade that has been automatically executed without the potential benefit of price improvement.',
    'Bunched Trade': 'Bunched Trade - The combining of multiple odd-lot or round-lot orders for the same security so that they can all be executed at the same time. All affected clients must agree to the bunching before the order is submitted. Bunched trades may also be referred to as block trades.',
    'Bunched Sold Trade': 'Bunched Sold Trade - A bunched trade that is reported late.',
    'Cash Sale': 'Cash Sale - A transaction which requires delivery of securities and payment on the same day the trade takes place.',
    'Closing Prints': 'Closing Prints - The Participant Closing Price represents the last qualifying trade paid for a security by a Participant during the trading day.',
    'Cross Trade': 'Cross Trade - Indicates that the trade resulted from a Market Center’s crossing session.',
    'Derivatively Priced': 'Derivatively Priced - The transaction that constituted the trade-through was the execution of an order at a price that was not based, directly or indirectly, on the quoted price of the NMS stock at the time of execution, and for which the material terms were not reasonably determinable at the time the commitment to execute the order was made (REG NMS 611b7).',
    'Distribution': 'Distribution stock refers to a large blocks of a security that are carefully sold into the market gradually in smaller blocks so as to inundate the market with sell orders for the security and driving down its price.',
    'Form T': 'Form T - Identifies a trade that was executed outside of regular primary market hours and is reported as an extended hours trade.',
    'Short Sale Restriction In Effect': 'The stock has declined by more than 10% in one trading session - which triggered the Short Sale Rule. This ticker cannot be shorted on a down-tick, and the SSR rule will be active for the day it triggered as well as the following day until the end of the day.',
    'Extended Trading Hours (Sold Out of Sequence)': 'Extended Hours (Sold out of Sequence) - Identifies a trade that takes place outside of regular market hours and is reported as an extended hours trade out of sequence and at a time different from the actual transaction time.',
    'Intermarket Sweep': 'A trade that occurs as a result of an order being routed to multiple markets to obtain the best price',
    'Market Center Official Close': 'Market Center Official Close - Indicates the ‘Official’ closing value as determined by a Market Center. This transaction report will contain the market center generated closing price.',
    'Market Center Official Open': 'Market Center Official Open - Indicates the ‘Official’ opening value as determined by a Market Center. This transaction report will contain the market center generated opening price.',
    'Market Center Opening Trade': 'Market Center Opening Trade - The trade that constituted the trade-through was a single priced opening transaction by the Market Center (REG NMS Rule 611b3).',
    'Market Center Reopening Trade': 'Market Center Reopening Trade - The trade that constituted the trade-through was a single priced reopening transaction by the Market Center (REG NMS Rule 611b3).',
    'Market Center Closing Trade': 'Market Center Closing Trade - The transaction that constituted the trade-through was a single priced closing transaction by the Market Center (REG NMS Rule 611b3).',
    'Next Day': 'Next Day - A transaction that requires the delivery of securities on the first business day following the trade date.',
    'Price Variation Trade': 'A trade that occurs as a result of a price change',
    'No Open No Resume': 'No Open No Resume - Indicates a trading halt or an opening delay is to be in effect for the rest of the trading day. Generally, a trading halt condition indicating the reason for the halt or delay is indicated prior to the input of Mode 16.',
    'Prior Reference Price': 'A trade that occurs at the prior day\'s closing price',
    'Rule 155 Trade (AMEX)': 'A trade that occurs as a result of an order being routed to multiple markets to obtain the best price',
    'Rule 127 NYSE': 'A trade that occurs as a result of an order being executed on the NYSE',
    'Opening Prints': 'Closing Prints - The trading days first drawings of a symbols candlestick charts.',
    'Stopped Stock (Regular Trade)': 'A trade that occurs after a stock has been halted',
    'Re-Opening Prints': 'A trade that occurs after a temporary halt in trading',
    'Seller': 'A Seller’s Option transaction gives the seller the right to deliver the security at any time within a specific period, ranging from not less than two calendar days, to not more than sixty calendar days. A security offered “Seller’s Option” may command a lesser price than if offered “Regular Way”.',
    'Sold Last': 'Sold Last sale condition modifier is used when a trade prints in sequence but is reported late OR the trade is printed by Amex in conformance to the One or Two Point Rule. A Sold Last transaction should only impact the consolidated last sale price for an issue if the market center reporting the sold last transaction also reported the transaction setting the current last sale price.',
    'Order Influx': 'Order Influx - A non-regulatory halt used when there is a severe influx of buy and sell orders. To prevent adisorderly market, trading is temporarily suspended.',
    'Sold (out of Sequence)': 'Sold Out of Sequence is used when a trade is printed (reported) out of sequence and at a time different from the actual transaction time.',
    'Split Trade': 'Split Trade - An execution in two markets when the specialist or Market Maker in the market first receiving the order agrees to execute a portion of it at whatever price is realized in another market to which the balance of the order is forwarded for execution.',
    'Stock Option Trade': 'Stock Option Trade - This is typically the stock portion of a delta neutral option trade executed by an option market maker.',
    'Yellow Flag Regular Trade': 'Yellow Flag Regular Trade - Market Centers will have the ability to identify regular trades being reported during specific events as out of the ordinary by appending a new sale condition code Yellow Flag (“Y”) on each transaction reported to the UTP SIP. The new sale condition “.Y” will be eligible to update all market center and consolidated statistics. In certain instances, the UTP SIP will be required to append the .Y for the market center for trades reported as regular-way (Sale Condition @)',
    'Odd Lot Trade': 'A trade that occurs for less than the normal trading unit',
    'Corrected Consolidated Close (per listing market)': 'Corrected Consolidated Close - A transaction executed by the Listing Market to establish the official Consolidated Last Price as indicated by the Listing Exchange.',
    'Held': 'Held - Trades received from a non-primary Participant during a primary market regulatory halt. These trades are held by the CTS Processor and are disseminated after the close of the primary market with an appropriate Held Trade Indicator code applicable to the trade.',
    'Trade Thru Exempt': 'The Trade Through rule is a 20 year-old rule applied to NYSE-listed stocks that states that when a market receives an order, it cannot execute it at a price inferior to any found on another market. In modern electronic markets where trades are executed in milliseconds, this rule can prevent a broker’s ability to meet their “best execution” obligation--because speed provides certainty that the price that is advertised can be accessed.',
    'Contingent Trade': 'Contengent Trade - A Sale Condition code used to identify a transaction where the execution of the transaction is contingent upon some event.',
    'Qualified Contingent Trade': 'Qualified Contingent Trade - A transaction consisting of two or more component orders executed as agent or principal where the execution of one component is contingent upon the execution of all other components at or near the same time and the price is determined by the relationship between the component orders and not the current market price for the security.',
    'Placeholder for 611 exempt': '611 Exempt - Rule 611, which is the Trade Through Exemption of SEC Regulation NMS, is very lengthy to cover in detail. Parties interested in reading the rule in its entirely should type "SEC Rule 611" into an internet search engine. This is the portion of the document that is pertinent to IB traders, in a nutshell: Typically the trades involved are a multi-component trade involving orders for a security and a related derivative, or, in the alternative, orders for related securities, that are executed at or near the same time. The SIA (Securities Industry Association) notes that the economics of a contingent trade are based on the relationship between the prices of the security and the related derivative or security, and that the execution of one order is contingent upon the execution of the other order.',
    'SSR in Effect': 'SSR - Any stock that has dropped more than 10% intraday has SSR in effect for that day and the following.',
    'Cash Only Settlement': 'Cash Only Settlement - The buyer of the option is not required to purchase or sell the underlying security',
    'Closing': 'Closing - The Participant Closing Price represents the last qualifying trade paid for a security by a Participant during the trading day.',
    'Slow Due Set Slow List Bid Ask': 'A trade that occurs as a result of a slow due set slow list bid and ask order',
    'Manual Bid and Ask': 'Manual Bid and Ask - Slow Quote Due to a NYSE Liquidity Replenishment Point (LRP), NYSE MKT Stock Market Tolerance Breach (Spread, Momentum or Gap Trade Tolerance), or Gap Quote on Both the Bid and Offer Sides',
    'Multi Leg auto-electronic trade': 'A trade that occurs as a result of a multi-leg auto-electronic order',
    'Single Leg Auction Non ISO': 'A trade that occurs as a result of a single-leg auction non-ISO order',
    'Multi Leg Auction': 'A trade that occurs as a result of a multi-leg auction order',
    'Automatic Execution': 'A trade that occurs automatically as a result of an algorithm or trading strategy'
}
OPTIONS_EXCHANGES = { 
    300:'NYSE American Options',
    301:'Boston Options Exchange',
    302:'Chicago Board Options Exchange',
    303:'MIAX Emerald, LLC',
    304:'Cboe EDGX Options',
    307:'Nasdaq Global Markets Exchange Group',
    308:'International Securities Exchange, LLC',
    309:'Nasdaq MRX Options Exchange',
    312:'MIAX International Securities Exchange, LLC',
    313:'NYSE Arca, Inc. - Options',
    314:'Options Price Reporting Authority',
    315:'MIAX Pearl, LLC - Options',
    316:'Nasdaq Options Market',
    319:'Nasdaq BX - Options',
    322:'Cboe C2 Options Exchange',
    323:'Nasdaq Philadelphia Exchange, LLC - Options',
    325:'Cboe BZX Options Exchange'}

STOCK_EXCHANGES = {
1:'NYSE American, LLC',
2:'Nasdaq OMX BX, Inc.',
3:'NYSE National, Inc.',
4:'FINRA NYSE TRF',
4:'FINRA Nasdaq TRF Carteret',
4:'FINRA Nasdaq TRF Chicago',
4:'FINRA Alternative Display Facility',
5:'Unlisted Trading Privileges',
6:'International Securities Exchange, LLC - Stocks',
7:'Cboe EDGA',
8:'Cboe EDGX',
9:'NYSE Chicago, Inc.',
10:'New York Stock Exchange',
11:'NYSE Arca, Inc.',
12:'Nasdaq',
13:'Consolidated Tape Association',
14:'Long-Term Stock Exchange',
15:'Investors Exchange',
16:'Cboe Stock Exchange',
17:'Nasdaq Philadelphia Exchange LLC',
18:'Cboe BYX',
19:'Cboe BZX',
20:'MIAX Pearl',
21:'Members Exchange',
62:'OTC Equity Security',    
 }

TAPES = {1: "NYSE", 2: "AMEX", 3: "Nasdaq"}


indicators = {
    50: "Normal",
    51: "Bankrupt",
    52: "Deficient - Below listing requirements",
    53: "Delinquent - Late filing",
    54: "Bankrupt and Deficient",
    55: "Bankrupt and Delinquent",
    56: "Deficient and Delinquent",
    57: "Deficient, Delinquent, and Bankrupt",
    58: "Liquidation",
    59: "Creations Suspended",
    60: "Redemptions Suspended",
    61: "Creations and/or Redemptions Suspended",
    100: "Normal Trading",
    101: "Opening Delay",
    102: "Trading Halt",
    103: "Resume",
    104: "No Open/No Resume",
    105: "Price Indication",
    106: "Trading Range Indication",
    107: "Market Imbalance Buy",
    108: "Market Imbalance Sell",
    109: "Market On Close Imbalance Buy",
    110: "Market On Close Imbalance Sell",
    111: "No Market Imbalance",
    112: "No Market On Close Imbalance",
    113: "Short Sale Restriction",
    114: "Limit Up-Limit Down",
    115: "Quotation Resumption",
    116: "Trading Resumption",
    117: "Volatility Trading Pause",
    300: "NONE",
    301: "Short Sales Restriction Activated",
    302: "Short Sales Restriction Continued",
    303: "Short Sales Restriction Deactivated",
    304: "Short Sales Restriction in Effect",
    305: "Short Sales Restriction Max",

    601: "NBBO_NO_CHANGE",
    602: "NBBO_QUOTE_IS_NBBO",
    603: "NBBO_NO_BB_NO_BO",
    604: "NBBO_BB_BO_SHORT_APPENDAGE",
    605: "NBBO_BB_BO_LONG_APPENDAGE",

    621: "HELD_TRADE_NOT_LAST_SALE_AND_NOT_ON_CONSOLIDATED",
    622: "HELD_TRADE_LAST_SALE_BUT_NOT_ON_CONSOLIDATED",
    623: "HELD_TRADE_LAST_SALE_AND_ON_CONSOLIDATED",


    501: "RETAIL_INTEREST_ON_BID NYSE Retail interest fact sheet",
    502: "RETAIL_INTEREST_ON_ASK NYSE Retail interest fact sheet",
    503: "RETAIL_INTEREST_ON_BID_AND_ASK NYSE Retail interest fact sheet",
    504: "FINRA_BBO_NO_CHANGE",
    505: "FINRA_BBO_DOES_NOT_EXIST",
    506: "FINRA_BB_BO_EXECUTABLE",
    507: "FINRA_BB_BELOW_LOWER_BAND",
    508: "FINRA_BO_ABOVE_UPPER_BAND",
    509: "FINRA_BB_BELOW_LOWER_BAND_BO_ABOVE_UPPER_BAND",
    901: "CTA_NOT_DUE_TO_RELATED_SECURITY",
    902: "CTA_DUE_TO_RELATED_SECURITY",
    903: "CTA_NOT_IN_VIEW_OF_COMMON",
    904: "CTA_IN_VIEW_OF_COMMON",
    905: "CTA_PRICE_INDICATOR",
    906: "CTA_NEW_PRICE_INDICATOR",
    907: "CTA_CORRECTED_PRICE_INDICATION",
    908: "CTA_CANCELLED_MARKET_IMBALANCE_PRICE_TRADING_RANGE_INDICATION",}


